# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

# pylint: disable=no-self-use

import json
import os
import os.path as osp
import shutil
from contextlib import contextmanager
from multiprocessing.pool import Pool
from typing import Dict, Optional

import numpy as np
import pycocotools.mask as mask_utils
from json_stream.writer import streamable_dict, streamable_list

from datumaro.components.annotation import (
    Annotation,
    Bbox,
    Caption,
    Cuboid3d,
    Ellipse,
    HashKey,
    Label,
    LabelCategories,
    Mask,
    MaskCategories,
    Points,
    PointsCategories,
    Polygon,
    PolyLine,
    RleMask,
    Shape,
)
from datumaro.components.crypter import NULL_CRYPTER
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.dataset_item_storage import ItemStatus
from datumaro.components.exporter import ExportContextComponent, Exporter
from datumaro.components.media import Image, MediaElement, PointCloud, Video, VideoFrame
from datumaro.util import cast, dump_json_file

from .format import DATUMARO_FORMAT_VERSION, DatumaroPath


class JsonWriter:
    @classmethod
    def _convert_attribute_categories(cls, attributes):
        return sorted(attributes)

    @classmethod
    def _convert_labels_label_groups(cls, labels):
        return sorted(labels)

    @classmethod
    def _convert_label_categories(cls, obj):
        converted = {
            "labels": [],
            "label_groups": [],
            "attributes": cls._convert_attribute_categories(obj.attributes),
        }
        for label in obj.items:
            converted["labels"].append(
                {
                    "name": cast(label.name, str),
                    "parent": cast(label.parent, str),
                    "attributes": cls._convert_attribute_categories(label.attributes),
                }
            )
        for label_group in obj.label_groups:
            converted["label_groups"].append(
                {
                    "name": cast(label_group.name, str),
                    "group_type": label_group.group_type.to_str(),
                    "labels": cls._convert_labels_label_groups(label_group.labels),
                }
            )
        return converted

    @classmethod
    def _convert_mask_categories(cls, obj):
        converted = {
            "colormap": [],
        }
        for label_id, color in obj.colormap.items():
            converted["colormap"].append(
                {
                    "label_id": int(label_id),
                    "r": int(color[0]),
                    "g": int(color[1]),
                    "b": int(color[2]),
                }
            )
        return converted

    @classmethod
    def _convert_points_categories(cls, obj):
        converted = {
            "items": [],
        }
        for label_id, item in obj.items.items():
            converted["items"].append(
                {
                    "label_id": int(label_id),
                    "labels": [cast(label, str) for label in item.labels],
                    "joints": [list(map(int, j)) for j in item.joints],
                }
            )
        return converted

    @classmethod
    def write_categories(cls, categories) -> Dict[str, Dict]:
        dict_cat = {}
        for ann_type, desc in categories.items():
            if isinstance(desc, LabelCategories):
                converted_desc = cls._convert_label_categories(desc)
            elif isinstance(desc, MaskCategories):
                converted_desc = cls._convert_mask_categories(desc)
            elif isinstance(desc, PointsCategories):
                converted_desc = cls._convert_points_categories(desc)
            else:
                raise NotImplementedError()
            dict_cat[ann_type.name] = converted_desc

        return dict_cat


class _SubsetWriter:
    def __init__(
        self,
        context: Exporter,
        subset: str,
        ann_file: str,
        export_context: ExportContextComponent,
    ):
        self._context = context
        self._subset = subset

        self._data = {
            "dm_format_version": DATUMARO_FORMAT_VERSION,
            "media_type": context._extractor.media_type()._type,
            "infos": {},
            "categories": {},
            "items": [],
        }

        self.ann_file = ann_file
        self.export_context = export_context

    @property
    def infos(self):
        return self._data["infos"]

    @property
    def categories(self):
        return self._data["categories"]

    @property
    def items(self):
        return self._data["items"]

    def is_empty(self):
        return not self.items

    @staticmethod
    @contextmanager
    def context_save_media(
        item: DatasetItem, context: ExportContextComponent, encryption: bool = False
    ) -> None:
        """Implicitly change the media path and save it if save_media=True.
        When done, revert it's path as before.

        Parameters
        ----------
            item: Dataset item to save its media
            context: Context instance to help the media export
            encryption: If false, prevent the media from being encrypted
        """
        if item.media is None:
            yield
        elif isinstance(item.media, Video):
            video = item.media_as(Video)

            if context.save_media:
                fname = context.make_video_filename(item)
                context.save_video(item, fname=fname, subdir=item.subset)
                item.media = Video(
                    path=fname,
                    step=video._step,
                    start_frame=video._start_frame,
                    end_frame=video._end_frame,
                )

            yield
            item.media = video
        elif isinstance(item.media, VideoFrame):
            video_frame = item.media_as(VideoFrame)

            if context.save_media:
                fname = context.make_video_filename(item)
                context.save_video(item, fname=fname, subdir=item.subset)
                item.media = VideoFrame(Video(fname), video_frame.index)

            yield
            item.media = video_frame
        elif isinstance(item.media, Image):
            image = item.media_as(Image)

            if context.save_media:
                # Temporarily update image path and save it.
                fname = context.make_image_filename(item)
                context.save_image(item, encryption=encryption, fname=fname, subdir=item.subset)
                item.media = Image.from_file(path=fname, size=image._size)

            yield
            item.media = image
        elif isinstance(item.media, PointCloud):
            pcd = item.media_as(PointCloud)

            if context.save_media:
                pcd_fname = context.make_pcd_filename(item)
                context.save_point_cloud(item, fname=pcd_fname, subdir=item.subset)

                extra_images = []
                for i, extra_image in enumerate(pcd.extra_images):
                    extra_images.append(
                        Image.from_file(
                            path=context.make_pcd_extra_image_filename(item, i, extra_image)
                        )
                    )

                # Temporarily update media with a new pcd saved into disk.
                item.media = PointCloud.from_file(path=pcd_fname, extra_images=extra_images)

            yield
            item.media = pcd
        else:
            raise NotImplementedError

    def add_item(self, item: DatasetItem, *args, **kwargs) -> None:
        self.items.append(self._gen_item_desc(item))

    def _gen_item_desc(self, item: DatasetItem, *args, **kwargs) -> Dict:
        annotations = []
        item_desc = {
            "id": item.id,
            "annotations": annotations,
        }

        if item.attributes:
            item_desc["attr"] = item.attributes

        with self.context_save_media(item, self.export_context):
            # Since VideoFrame is a descendant of Image, this condition should be ahead of Image
            if isinstance(item.media, VideoFrame):
                video_frame = item.media_as(VideoFrame)
                item_desc["video_frame"] = {
                    "video_path": getattr(video_frame.video, "path", None),
                    "frame_index": getattr(video_frame, "index", -1),
                }
            elif isinstance(item.media, Video):
                video = item.media_as(Video)
                item_desc["video"] = {
                    "path": getattr(video, "path", None),
                    "step": video._step,
                    "start_frame": video._start_frame,
                }
                if video._end_frame is not None:
                    item_desc["video"]["end_frame"] = video._end_frame
            elif isinstance(item.media, Image):
                image = item.media_as(Image)
                item_desc["image"] = {"path": getattr(image, "path", None)}
                if item.media.has_size:  # avoid occasional loading
                    item_desc["image"]["size"] = image.size
            elif isinstance(item.media, PointCloud):
                pcd = item.media_as(PointCloud)

                item_desc["point_cloud"] = {"path": getattr(pcd, "path", None)}

                related_images = [
                    {"path": getattr(img, "path", None), "size": img.size}
                    if img.has_size
                    else {"path": getattr(img, "path", None)}
                    for img in pcd.extra_images
                ]

                if related_images:
                    item_desc["related_images"] = related_images
            elif isinstance(item.media, MediaElement):
                item_desc["media"] = {"path": getattr(item.media, "path", None)}

        for ann in item.annotations:
            if isinstance(ann, Label):
                converted_ann = self._convert_label_object(ann)
            elif isinstance(ann, Mask):
                converted_ann = self._convert_mask_object(ann)
            elif isinstance(ann, Points):
                converted_ann = self._convert_points_object(ann)
            elif isinstance(ann, PolyLine):
                converted_ann = self._convert_polyline_object(ann)
            elif isinstance(ann, Polygon):
                converted_ann = self._convert_polygon_object(ann)
            elif isinstance(ann, Bbox):
                converted_ann = self._convert_bbox_object(ann)
            elif isinstance(ann, Caption):
                converted_ann = self._convert_caption_object(ann)
            elif isinstance(ann, Cuboid3d):
                converted_ann = self._convert_cuboid_3d_object(ann)
            elif isinstance(ann, Ellipse):
                converted_ann = self._convert_ellipse_object(ann)
            elif isinstance(ann, HashKey):
                continue
            else:
                raise NotImplementedError()
            annotations.append(converted_ann)

        return item_desc

    def add_infos(self, infos):
        self._data["infos"].update(infos)

    def add_categories(self, categories):
        self._data["categories"] = JsonWriter.write_categories(categories)

    def write(self, *args, **kwargs):
        dump_json_file(self.ann_file, self._data)

    def _convert_annotation(self, obj):
        assert isinstance(obj, Annotation)

        ann_json = {
            "id": cast(obj.id, int),
            "type": cast(obj.type.name, str),
            "attributes": obj.attributes,
            "group": cast(obj.group, int, 0),
        }
        if obj.object_id >= 0:
            ann_json["object_id"] = cast(obj.object_id, int)
        return ann_json

    def _convert_label_object(self, obj):
        converted = self._convert_annotation(obj)

        converted.update(
            {
                "label_id": cast(obj.label, int),
            }
        )
        return converted

    def _convert_mask_object(self, obj):
        converted = self._convert_annotation(obj)

        if isinstance(obj, RleMask):
            rle = obj.rle
        else:
            rle = mask_utils.encode(np.require(obj.image, dtype=np.uint8, requirements="F"))

        if isinstance(rle["counts"], str):
            counts = rle["counts"]
        else:
            counts = rle["counts"].decode("ascii")

        converted.update(
            {
                "label_id": cast(obj.label, int),
                "rle": {
                    # serialize as compressed COCO mask
                    "counts": counts,
                    "size": list(int(c) for c in rle["size"]),
                },
                "z_order": obj.z_order,
            }
        )
        return converted

    def _convert_shape_object(self, obj):
        assert isinstance(obj, Shape)
        converted = self._convert_annotation(obj)

        converted.update(
            {
                "label_id": cast(obj.label, int),
                "points": [float(p) for p in obj.points],
                "z_order": obj.z_order,
            }
        )
        return converted

    def _convert_polyline_object(self, obj):
        return self._convert_shape_object(obj)

    def _convert_polygon_object(self, obj):
        return self._convert_shape_object(obj)

    def _convert_bbox_object(self, obj):
        converted = self._convert_shape_object(obj)
        converted.pop("points", None)
        converted["bbox"] = [float(p) for p in obj.get_bbox()]
        return converted

    def _convert_points_object(self, obj):
        converted = self._convert_shape_object(obj)

        converted.update(
            {
                "visibility": [int(v.value) for v in obj.visibility],
            }
        )
        return converted

    def _convert_caption_object(self, obj):
        converted = self._convert_annotation(obj)

        converted.update(
            {
                "caption": cast(obj.caption, str),
            }
        )
        return converted

    def _convert_cuboid_3d_object(self, obj):
        converted = self._convert_annotation(obj)
        converted.update(
            {
                "label_id": cast(obj.label, int),
                "position": [float(p) for p in obj.position],
                "rotation": [float(p) for p in obj.rotation],
                "scale": [float(p) for p in obj.scale],
            }
        )
        return converted

    def _convert_ellipse_object(self, obj: Ellipse):
        return self._convert_shape_object(obj)


class _StreamSubsetWriter(_SubsetWriter):
    def __init__(
        self,
        context: Exporter,
        subset: str,
        ann_file: str,
        export_context: ExportContextComponent,
    ):
        super().__init__(context, subset, ann_file, export_context)

    def write(self, *args, **kwargs):
        @streamable_list
        def _item_list():
            subset = self._context._extractor.get_subset(self._subset)
            pbar = self._context._ctx.progress_reporter
            for item in pbar.iter(subset, desc=f"Exporting '{self._subset}'"):
                yield self._gen_item_desc(item)

        @streamable_dict
        def _data():
            yield "dm_format_version", self._data["dm_format_version"]
            yield "media_type", self._data["media_type"]
            yield "infos", self._data["infos"]
            yield "categories", self._data["categories"]
            yield "items", _item_list()

        with open(self.ann_file, "w", encoding="utf-8") as fp:
            json.dump(_data(), fp)

    def is_empty(self):
        # TODO: Force empty to be False, but it should be fixed with refactoring `_SubsetWriter`.`
        return False


class DatumaroExporter(Exporter):
    DEFAULT_IMAGE_EXT = DatumaroPath.IMAGE_EXT
    PATH_CLS = DatumaroPath

    def create_writer(
        self,
        subset: str,
        images_dir: str,
        pcd_dir: str,
        video_dir: str,
    ) -> _SubsetWriter:
        export_context = ExportContextComponent(
            save_dir=self._save_dir,
            save_media=self._save_media,
            images_dir=images_dir,
            pcd_dir=pcd_dir,
            video_dir=video_dir,
            crypter=NULL_CRYPTER,
            image_ext=self._image_ext,
            default_image_ext=self._default_image_ext,
        )

        return (
            _SubsetWriter(
                context=self,
                subset=subset,
                ann_file=osp.join(self._annotations_dir, subset + self.PATH_CLS.ANNOTATION_EXT),
                export_context=export_context,
            )
            if not self._stream
            else _StreamSubsetWriter(
                context=self,
                subset=subset,
                ann_file=osp.join(self._annotations_dir, subset + self.PATH_CLS.ANNOTATION_EXT),
                export_context=export_context,
            )
        )

    def _apply_impl(self, pool: Optional[Pool] = None, *args, **kwargs):
        os.makedirs(self._save_dir, exist_ok=True)

        images_dir = osp.join(self._save_dir, self.PATH_CLS.IMAGES_DIR)
        os.makedirs(images_dir, exist_ok=True)
        self._images_dir = images_dir

        annotations_dir = osp.join(self._save_dir, self.PATH_CLS.ANNOTATIONS_DIR)
        os.makedirs(annotations_dir, exist_ok=True)
        self._annotations_dir = annotations_dir

        self._pcd_dir = osp.join(self._save_dir, self.PATH_CLS.PCD_DIR)
        self._video_dir = osp.join(self._save_dir, self.PATH_CLS.VIDEO_DIR)

        writers = {
            subset: self.create_writer(
                subset,
                self._images_dir,
                self._pcd_dir,
                self._video_dir,
            )
            for subset in self._extractor.subsets()
        }

        for writer in writers.values():
            writer.add_infos(self._extractor.infos())
            writer.add_categories(self._extractor.categories())

        pbar = self._ctx.progress_reporter
        for subset_name, subset in self._extractor.subsets().items():
            if not self._stream:
                for item in pbar.iter(subset, desc=f"Exporting '{subset_name}'"):
                    writers[subset_name].add_item(item, pool)

        for subset, writer in writers.items():
            if self._patch and subset in self._patch.updated_subsets and writer.is_empty():
                if osp.isfile(writer.ann_file):
                    # Remove subsets that became empty
                    os.remove(writer.ann_file)
                continue

            writer.write(pool)

        if self._save_hashkey_meta:
            self._save_hashkey_file(self._save_dir)

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        for subset in patch.updated_subsets:
            conv = cls(dataset.get_subset(subset), save_dir=save_dir, **kwargs)
            conv._patch = patch
            conv.apply()

        conv = cls(dataset, save_dir=save_dir, **kwargs)
        for (item_id, subset), status in patch.updated_items.items():
            if status != ItemStatus.removed:
                item = patch.data.get(item_id, subset)
            else:
                item = DatasetItem(item_id, subset=subset)

            if not (status == ItemStatus.removed or not item.media):
                continue

            image_path = osp.join(
                save_dir, cls.PATH_CLS.IMAGES_DIR, item.subset, conv._make_image_filename(item)
            )
            if osp.isfile(image_path):
                os.unlink(image_path)

            pcd_path = osp.join(
                save_dir, cls.PATH_CLS.PCD_DIR, item.subset, conv._make_pcd_filename(item)
            )
            if osp.isfile(pcd_path):
                os.unlink(pcd_path)

            related_images_path = osp.join(save_dir, cls.PATH_CLS.IMAGES_DIR, item.subset, item.id)
            if osp.isdir(related_images_path):
                shutil.rmtree(related_images_path)

    @property
    def can_stream(self) -> bool:
        return True



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
from typing import Dict, List, Optional

from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.errors import DatasetImportError
from datumaro.rust_api import JsonSectionPageMapper
from datumaro.util import parse_json


class SegmentAnythingImporter(Importer):
    _N_JSON_TO_TEST = 10
    _MAX_ANNOTATION_SECTION_BYTES = 100 * 1024 * 1024  # 100 MiB
    _ANNO_EXT = ".json"

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> Optional[FormatDetectionConfidence]:
        # test maximum 10 annotation files only
        ctr = 0
        for file in context.require_files_iter(f"*{cls._ANNO_EXT}"):
            ctr += 1
            with context.probe_text_file(
                file, "Annotation format is not Segmentat-Anything format", is_binary_file=True
            ) as f:
                fpath = os.path.join(context.root_path, file)
                page_mapper = JsonSectionPageMapper(fpath)
                sections = page_mapper.sections()

                if set(sections.keys()) != {"annotations", "image"}:
                    raise DatasetImportError

                offset, size = sections["image"]["offset"], sections["image"]["size"]
                f.seek(offset, 0)
                img_contents = parse_json(f.read(size))

                if set(img_contents.keys()) != {
                    "image_id",
                    "width",
                    "height",
                    "file_name",
                }:
                    raise DatasetImportError

                offset, size = sections["annotations"]["offset"], sections["annotations"]["size"]

                if size > cls._MAX_ANNOTATION_SECTION_BYTES:
                    msg = f"Annotation section is too huge. It exceeded {cls._MAX_ANNOTATION_SECTION_BYTES} bytes."
                    raise DatasetImportError(msg)

                f.seek(offset, 0)
                ann_contents = parse_json(f.read(size))

                if not {"id", "segmentation", "bbox"}.issubset(set(ann_contents[0])):
                    raise DatasetImportError

            if ctr > cls._N_JSON_TO_TEST:
                break

    @classmethod
    def find_sources(cls, path) -> List[Dict]:
        if not os.path.isdir(path):
            return []
        return [{"url": path, "format": cls.NAME}]

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
from typing import List, Type

from datumaro.cli.util import MultilineFormatter
from datumaro.util import to_snake_case

_plugin_types = None


def plugin_types() -> List[Type["CliPlugin"]]:
    global _plugin_types
    if _plugin_types is None:
        from datumaro.components.dataset_base import DatasetBase
        from datumaro.components.exporter import Exporter
        from datumaro.components.importer import Importer
        from datumaro.components.launcher import Launcher
        from datumaro.components.transformer import Transform
        from datumaro.components.validator import Validator

        _plugin_types = [Launcher, DatasetBase, Transform, Importer, Exporter, Validator]

    return _plugin_types


def remove_plugin_type(s):
    for t in {"transform", "base", "exporter", "launcher", "importer", "validator"}:
        s = s.replace("_" + t, "")
    return s


class _PluginNameDescriptor:
    def __get__(self, obj, objtype=None):
        if not objtype:
            objtype = type(obj)
        return remove_plugin_type(to_snake_case(objtype.__name__))


class CliPlugin:
    NAME = _PluginNameDescriptor()

    @staticmethod
    def _get_doc(cls):
        doc = getattr(cls, "__doc__", "")
        if doc:
            if any(getattr(t, "__doc__", "") == doc for t in plugin_types()):
                doc = ""
        return doc

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        args = {
            "prog": cls.NAME,
            "description": cls._get_doc(cls),
            "formatter_class": MultilineFormatter,
        }
        args.update(kwargs)

        return argparse.ArgumentParser(**args)

    @classmethod
    def parse_cmdline(cls, args=None):
        if args and args[0] == "--":
            args = args[1:]
        parser = cls.build_cmdline_parser()
        args = parser.parse_args(args)
        args = vars(args)

        log.debug(
            "Parsed parameters: \n\t%s", "\n\t".join("%s: %s" % (k, v) for k, v in args.items())
        )

        return args



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import Dict, Optional, Tuple

from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import MediaType
from datumaro.plugins.data_formats.datumaro_binary.mapper.annotation import AnnotationListMapper

from .common import DictMapper, Mapper, StringMapper
from .media import MediaMapper


class DatasetItemMapper(Mapper):
    @staticmethod
    def forward(obj: DatasetItem) -> bytes:
        bytes_arr = bytearray()
        bytes_arr.extend(StringMapper.forward(obj.id))
        bytes_arr.extend(StringMapper.forward(obj.subset))
        bytes_arr.extend(MediaMapper.forward(obj.media))
        bytes_arr.extend(DictMapper.forward(obj.attributes))
        bytes_arr.extend(AnnotationListMapper.forward(obj.annotations))
        return bytes(bytes_arr)

    @staticmethod
    def backward(
        _bytes: bytes, offset: int = 0, media_path_prefix: Optional[Dict[MediaType, str]] = None
    ) -> Tuple[DatasetItem, int]:
        id, offset = StringMapper.backward(_bytes, offset)
        subset, offset = StringMapper.backward(_bytes, offset)
        media, offset = MediaMapper.backward(_bytes, offset, media_path_prefix)
        attributes, offset = DictMapper.backward(_bytes, offset)
        annotations, offset = AnnotationListMapper.backward(_bytes, offset)
        return (
            DatasetItem(
                id=id, subset=subset, media=media, attributes=attributes, annotations=annotations
            ),
            offset,
        )



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging as log
import os
import os.path as osp
from typing import Any, Dict, List, Optional, Type, TypeVar

from datumaro.components.annotation import Annotation, AnnotationType, LabelCategories, RotatedBbox
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME, DatasetItem, SubsetBase
from datumaro.components.errors import (
    DatasetExportError,
    DatasetImportError,
    InvalidAnnotationError,
    MediaTypeError,
)
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.image import IMAGE_EXTENSIONS
from datumaro.util.os_util import find_files

T = TypeVar("T")


class DotaFormat:
    ANNOTATION_DIR = "labelTxt"
    IMAGE_DIR = "images"


class DotaBase(SubsetBase):
    def __init__(
        self,
        path: Optional[List[str]] = None,
        *,
        img_path: Optional[str] = None,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ) -> None:
        super().__init__(subset=subset, ctx=ctx)

        if not osp.isdir(path):
            raise DatasetImportError(f"Can't find annotation directory {path}")

        self._path = path

        super().__init__(subset=subset, ctx=ctx)

        self._img_files = self._load_img_files(img_path)
        self._label_categories = self._load_categories(path)
        self._categories = {AnnotationType.label: self._label_categories}

        self._items = self._load_items(path)

    def _load_img_files(self, rootpath: str) -> Dict[str, str]:
        return {
            self._get_fname(img_file): img_file
            for img_file in find_files(rootpath, IMAGE_EXTENSIONS, recursive=True, max_depth=2)
        }

    def _load_categories(self, path):
        label_names = []
        for ann_file in os.listdir(path):
            label_names.extend(
                self._parse_annotations(
                    ann_file=osp.join(self._path, ann_file), only_label_names=True
                )
            )

        label_categories = LabelCategories()
        for label_name in sorted(set(label_names)):
            label_categories.add(label_name)

        return label_categories

    def _load_items(self, path):
        items = []
        for ann_file in os.listdir(path):
            fname = osp.splitext(ann_file)[0]
            img = Image.from_file(path=self._img_files[fname])
            anns = self._parse_annotations(
                ann_file=osp.join(self._path, ann_file), only_label_names=False
            )
            items.append(DatasetItem(id=fname, subset=self._subset, media=img, annotations=anns))
        return items

    def _get_fname(self, fpath: str) -> str:
        return osp.splitext(osp.basename(fpath))[0]

    def _parse_annotations(
        self,
        ann_file: str,
        only_label_names: bool,
    ) -> List[Annotation]:
        lines = []
        with open(ann_file, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if line:
                    lines.append(line)

        annotations = []
        for line in lines:
            parts = line.split()
            if len(parts) != 10:
                log.debug(
                    f"Unexpected field count {len(parts)} in the bbox description. "
                    "Expected 10 fields (8 coordinates for rectangle, category, and difficulty)."
                )
                continue

            label_name = self._parse_field(parts[-2], str, "label_name")

            if only_label_names:
                annotations.append(label_name)
                continue

            label_id, _ = self._label_categories.find(label_name)
            coords = [
                (
                    self._parse_field(parts[i], float, "coords"),
                    self._parse_field(parts[i + 1], float, "coords"),
                )
                for i in range(0, 8, 2)
            ]
            difficulty = self._parse_field(parts[-1], int, "difficulty")

            annotations.append(
                RotatedBbox.from_rectangle(
                    coords, label=label_id, attributes={"difficulty": difficulty}
                )
            )
            self._ann_types.add(AnnotationType.rotated_bbox)

        return annotations

    @staticmethod
    def _parse_field(value: str, desired_type: Type[T], field_name: str) -> T:
        try:
            return desired_type(value)
        except Exception as e:
            raise InvalidAnnotationError(
                f"Can't parse {field_name} from '{value}'. Expected {desired_type}"
            ) from e


class DotaImporter(Importer):
    _ANNO_EXT = ".txt"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        context.require_file("**/" + DotaFormat.ANNOTATION_DIR + "/*" + cls._ANNO_EXT)
        return FormatDetectionConfidence.MEDIUM

    @classmethod
    def find_sources(cls, path: str) -> List[Dict[str, Any]]:
        sources = cls._find_sources_recursive(
            path=path,
            ext=cls._ANNO_EXT,
            dirname=DotaFormat.ANNOTATION_DIR,
            extractor_name="dota",
        )

        data_paths = set()
        for source in sources:
            url = osp.dirname(source["url"])
            subset_name = osp.relpath(source["url"], path).split(osp.sep)[0]
            data_paths.add((subset_name, url))

        return [
            {
                "url": ann_dir,
                "format": "dota",
                "options": {
                    "subset": subset,
                    "img_path": osp.join(path, subset, DotaFormat.IMAGE_DIR),
                },
            }
            for subset, ann_dir in data_paths
        ]

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]


class DotaExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".png"

    def _apply_impl(self):
        extractor = self._extractor
        save_dir = self._save_dir

        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        os.makedirs(save_dir, exist_ok=True)

        label_categories = extractor.categories()[AnnotationType.label]

        subsets = self._extractor.subsets()
        for subset_name, subset in subsets.items():
            if not subset_name or subset_name == DEFAULT_SUBSET_NAME:
                subset_name = DEFAULT_SUBSET_NAME

            subset_dir = osp.join(save_dir, subset_name)
            os.makedirs(subset_dir, exist_ok=True)

            for item in subset:
                try:
                    self._export_media(item, subset_dir)
                    self._export_item_annotation(item, subset_dir, label_categories)

                except Exception as e:
                    self._ctx.error_policy.report_item_error(e, item_id=(item.id, item.subset))

    def _export_media(self, item: DatasetItem, subset_dir: str) -> str:
        try:
            if not item.media or not (item.media.has_data or item.media.has_size):
                raise DatasetExportError(
                    "Failed to export item '%s': " "item has no image info" % item.id
                )

            image_name = self._make_image_filename(item)
            image_fpath = osp.join(subset_dir, DotaFormat.IMAGE_DIR, image_name)

            if self._save_media:
                self._save_image(item, image_fpath)

        except Exception as e:
            self._ctx.error_policy.report_item_error(e, item_id=(item.id, item.subset))

    def _export_item_annotation(
        self, item: DatasetItem, subset_dir: str, label_categories: LabelCategories
    ) -> None:
        try:
            annotations = ""
            for bbox in item.annotations:
                if not isinstance(bbox, RotatedBbox) or bbox.label is None:
                    continue
                coords = bbox.as_polygon()
                coords = " ".join("%.2f %.2f" % (x, y) for x, y in coords)
                label_name = label_categories[bbox.label].name
                difficulty = bbox.attributes.get("difficulty", 0)
                annotations += "%s %s %s\n" % (coords, label_name, difficulty)

            annotation_path = osp.join(subset_dir, DotaFormat.ANNOTATION_DIR, "%s.txt" % item.id)
            os.makedirs(osp.dirname(annotation_path), exist_ok=True)

            with open(annotation_path, "w", encoding="utf-8") as f:
                f.write(annotations)

        except Exception as e:
            self._ctx.error_policy.report_item_error(e, item_id=(item.id, item.subset))



# Copyright (C) 2020-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import json
import logging as log
import os
import os.path as osp
from dataclasses import dataclass
from enum import Enum, auto
from io import BufferedWriter
from itertools import chain, groupby
from typing import Dict, List, Optional, Type, Union

import pycocotools.mask as mask_utils
from json_stream.writer import streamable_dict, streamable_list

import datumaro.util.annotation_util as anno_tools
import datumaro.util.mask_tools as mask_tools
from datumaro.components.annotation import (
    COORDINATE_ROUNDING_DIGITS,
    AnnotationType,
    Ellipse,
    Points,
    Polygon,
)
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.dataset_item_storage import ItemStatus
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.media import Image
from datumaro.util import cast, dump_json, dump_json_file, find, parse_json, str_to_bool
from datumaro.util.image import save_image

from .format import CocoPath, CocoTask


class SegmentationMode(Enum):
    guess = auto()
    polygons = auto()
    mask = auto()


@dataclass
class _Writer:
    fp: BufferedWriter
    is_empty: bool


class TemporaryWriters:
    def __init__(self, subset: str, task: CocoTask, ann_dir: str):
        self._subset = subset
        self._task = task
        self._ann_dir = ann_dir
        self._writers = tuple()
        self.reset()

    def close(self):
        for writer in self._writers:
            if not writer.fp.closed:
                writer.fp.close()

    def remove(self):
        self.close()

        for writer in self._writers:
            fpath = writer.fp.name
            if osp.exists(fpath):
                os.remove(fpath)

    def reset(self):
        self.remove()

        self._writers = tuple(
            _Writer(
                fp=open(
                    osp.join(self._ann_dir, f"__{self._task.name}_{self._subset}_{key}.tmp"), "wb"
                ),
                is_empty=True,
            )
            for key in ["imgs", "anns"]
        )

    def __del__(self):
        self.remove()

    @property
    def imgs(self) -> _Writer:
        return self._writers[0]

    @property
    def anns(self) -> _Writer:
        return self._writers[1]

    def add_item(self, data: Dict) -> None:
        self.imgs.is_empty = False
        writer = self.imgs.fp
        writer.write(dump_json(data, append_newline=True))

    def add_anns(self, data: Dict) -> None:
        self.anns.is_empty = False
        writer = self.anns.fp
        writer.write(dump_json(data, append_newline=True))

    def merge(self, path: str, header: Dict, min_ann_id: Optional[int]) -> None:
        self.close()

        @streamable_list
        def _gen_images():
            with open(self.imgs.fp.name, "rb") as fp:
                for line in fp:
                    yield parse_json(line)

        @streamable_list
        def _gen_anns():
            with open(self.anns.fp.name, "rb") as fp:
                next_id = min_ann_id
                for line in fp:
                    ann = parse_json(line)
                    if min_ann_id is not None and not ann["id"]:
                        ann["id"] = next_id
                        next_id += 1
                    yield ann

        @streamable_dict
        def _gen():
            yield "licenses", header["licenses"]
            yield "info", header["info"]
            yield "categories", header["categories"]

            if not self.imgs.is_empty:
                yield "images", _gen_images()
            else:
                yield "images", []

            if not self.anns.is_empty:
                yield "annotations", _gen_anns()
            else:
                yield "annotations", []

        with open(path, "w", encoding="utf-8") as fp:
            json.dump(_gen(), fp)

        self.remove()


class _TaskExporter:
    def __init__(
        self, context: "CocoExporter", subset: str, task: CocoTask, ann_dir: str, stream: bool
    ):
        self._min_ann_id = 1
        self._context = context
        self._subset = subset
        self._task = task
        self._ann_dir = ann_dir
        self._stream = stream

        data = {"licenses": [], "info": {}, "categories": [], "images": [], "annotations": []}

        data["licenses"].append({"name": "", "id": 0, "url": ""})

        data["info"] = {
            "contributor": "",
            "date_created": "",
            "description": "",
            "url": "",
            "version": "",
            "year": "",
        }
        self._data = data
        self._temporary_writers = TemporaryWriters(
            subset=subset,
            task=task,
            ann_dir=ann_dir,
        )

    def is_empty(self):
        return (
            len(self._data["annotations"]) == 0
            if not self._stream
            else self._temporary_writers.anns.is_empty
        )

    def _get_image_id(self, item):
        return self._context._get_image_id(item)

    def save_image_info(self, item, filename):
        h = 0
        w = 0
        if item.media and item.media.size:
            h, w = item.media.size

        item_desc = {
            "id": self._get_image_id(item),
            "width": int(w),
            "height": int(h),
            "file_name": cast(filename, str, ""),
            "license": 0,
            "flickr_url": "",
            "coco_url": "",
            "date_captured": 0,
        }

        if not self._stream:
            self._data["images"].append(item_desc)
        else:
            self._temporary_writers.add_item(item_desc)

    def save_categories(self, dataset):
        raise NotImplementedError()

    def save_annotations(self, item):
        raise NotImplementedError()

    def write(self, path):
        next_id = self._min_ann_id
        for ann in self.annotations:
            if not ann["id"]:
                ann["id"] = next_id
                next_id += 1

        dump_json_file(path, self._data)

        if self._stream:
            self._temporary_writers.merge(path, self._data, self._min_ann_id)

    @property
    def annotations(self):
        return self._data["annotations"]

    @property
    def infos(self):
        return self._data["info"]

    @property
    def categories(self):
        return self._data["categories"]

    def _get_ann_id(self, annotation):
        ann_id = 0 if self._context._reindex else annotation.id
        if ann_id:
            self._min_ann_id = max(ann_id, self._min_ann_id)
        return ann_id

    @staticmethod
    def _convert_attributes(ann):
        return {k: v for k, v in ann.attributes.items() if k not in {"is_crowd", "score"}}


class _ImageInfoExporter(_TaskExporter):
    def is_empty(self):
        return (
            len(self._data["images"]) == 0
            if not self._stream
            else self._temporary_writers.imgs.is_empty
        )

    def save_categories(self, dataset):
        pass

    def save_annotations(self, item):
        pass


class _CaptionsExporter(_TaskExporter):
    def save_categories(self, dataset):
        pass

    def save_annotations(self, item):
        for ann_idx, ann in enumerate(item.annotations):
            if ann.type != AnnotationType.caption:
                continue

            elem = {
                "id": self._get_ann_id(ann),
                "image_id": self._get_image_id(item),
                "category_id": 0,  # NOTE: workaround for a bug in cocoapi
                "caption": ann.caption,
            }
            if "score" in ann.attributes:
                try:
                    elem["score"] = float(ann.attributes["score"])
                except Exception as e:
                    log.warning(
                        "Item '%s', ann #%s: failed to convert "
                        "attribute 'score': %e" % (item.id, ann_idx, e)
                    )
            if self._context._allow_attributes:
                attrs = self._convert_attributes(ann)
                if attrs:
                    elem["attributes"] = attrs

            if not self._stream:
                self.annotations.append(elem)
            else:
                self._temporary_writers.add_anns(elem)


class _InstancesExporter(_TaskExporter):
    _polygon_types = {AnnotationType.polygon, AnnotationType.ellipse}
    _allowed_types = {
        AnnotationType.bbox,
        AnnotationType.polygon,
        AnnotationType.mask,
        AnnotationType.ellipse,
    }

    def save_categories(self, dataset):
        label_categories = dataset.categories().get(AnnotationType.label)
        if label_categories is None:
            return

        for idx, cat in enumerate(label_categories.items):
            self.categories.append(
                {
                    "id": 1 + idx,
                    "name": cast(cat.name, str, ""),
                    "supercategory": cast(cat.parent, str, ""),
                }
            )

    @classmethod
    def crop_segments(cls, instances, img_width, img_height):
        instances = sorted(instances, key=lambda x: x[0].z_order)

        segment_map = []
        segments = []
        for inst_idx, (_, polygons, mask, _) in enumerate(instances):
            if polygons:
                segment_map.extend(inst_idx for p in polygons)
                segments.extend(polygons)
            elif mask is not None:
                segment_map.append(inst_idx)
                segments.append(mask)

        segments = mask_tools.crop_covered_segments(segments, img_width, img_height)

        for inst_idx, inst in enumerate(instances):
            new_segments = [s for si_id, s in zip(segment_map, segments) if si_id == inst_idx]

            if not new_segments:
                inst[1] = []
                inst[2] = None
                continue

            if inst[1]:
                inst[1] = sum(new_segments, [])
            else:
                mask = mask_tools.merge_masks(new_segments)
                inst[2] = mask_tools.mask_to_rle(mask)

        return instances

    def find_instance_parts(self, group, img_width, img_height):
        boxes = [a for a in group if a.type == AnnotationType.bbox]
        polygons: List[Union[Polygon, Ellipse]] = [
            a for a in group if a.type in self._polygon_types
        ]
        masks = [a for a in group if a.type == AnnotationType.mask]

        anns = boxes + polygons + masks
        leader = anno_tools.find_group_leader(anns)
        if len(boxes) > 0:
            bbox = anno_tools.max_bbox(boxes)
        else:
            bbox = anno_tools.max_bbox(anns)
        mask = None
        polygons = [p.as_polygon() for p in polygons]

        if self._context._segmentation_mode == SegmentationMode.guess:
            use_masks = True is leader.attributes.get(
                "is_crowd", find(masks, lambda x: x.label == leader.label) is not None
            )
        elif self._context._segmentation_mode == SegmentationMode.polygons:
            use_masks = False
        elif self._context._segmentation_mode == SegmentationMode.mask:
            use_masks = True
        else:
            raise NotImplementedError(
                "Unexpected segmentation mode '%s'" % self._context._segmentation_mode
            )

        if use_masks:
            if polygons and img_width > 0 and img_height > 0:
                mask = mask_tools.rles_to_mask(polygons, img_width, img_height)

            if masks:
                masks = (m.image for m in masks)
                if mask is not None:
                    masks = chain(masks, [mask])
                mask = mask_tools.merge_masks(masks)

            if mask is not None:
                mask = mask_tools.mask_to_rle(mask)
            polygons = []
        else:
            if masks:
                mask = mask_tools.merge_masks(m.image for m in masks)
                polygons += mask_tools.mask_to_polygons(mask)
            mask = None

        return [leader, polygons, mask, bbox]

    @staticmethod
    def find_instance_anns(annotations):
        return [a for a in annotations if a.type in _InstancesExporter._allowed_types]

    @classmethod
    def find_instances(cls, annotations):
        return anno_tools.find_instances(cls.find_instance_anns(annotations))

    def save_annotations(self, item):
        instances = self.find_instances(item.annotations)
        if not instances:
            return

        if not item.media or not item.media.size:
            h, w = 0, 0
            log.warning(
                "Item '%s': Mask annotations can be skipped since no image info available" % item.id
            )
        else:
            h, w = item.media.size

        instances = [self.find_instance_parts(i, w, h) for i in instances]

        if self._context._crop_covered and w > 0 and h > 0:
            instances = self.crop_segments(instances, w, h)

        for instance in instances:
            elem = self.convert_instance(instance, item)

            if elem is None:
                continue

            if not self._stream:
                self.annotations.append(elem)
            else:
                self._temporary_writers.add_anns(elem)

    def convert_instance(self, instance, item) -> Optional[Dict]:
        ann, polygons, mask, bbox = instance

        is_crowd = mask is not None
        if is_crowd:
            segmentation = {
                "counts": list(int(c) for c in mask["counts"]),
                "size": list(int(c) for c in mask["size"]),
            }
        else:
            segmentation = [list(map(float, p)) for p in polygons]

        area = 0
        if segmentation:
            if item.media and item.media.size:
                h, w = item.media.size
            else:
                # Here we can guess the image size as
                # it is only needed for the area computation
                w = bbox[0] + bbox[2]
                h = bbox[1] + bbox[3]

            rles = mask_utils.frPyObjects(segmentation, h, w)
            if is_crowd:
                rles = [rles]
            else:
                rles = mask_utils.merge(rles)
            area = mask_utils.area(rles)
        else:
            _, _, w, h = bbox
            segmentation = []
            area = w * h

        elem = {
            "id": self._get_ann_id(ann),
            "image_id": self._get_image_id(item),
            "category_id": cast(ann.label, int, -1) + 1,
            "segmentation": segmentation,
            "area": float(area),
            "bbox": [round(float(n), COORDINATE_ROUNDING_DIGITS) for n in bbox],
            "iscrowd": int(is_crowd),
        }
        if "score" in ann.attributes:
            try:
                elem["score"] = float(ann.attributes["score"])
            except Exception as e:
                log.warning("Item '%s': failed to convert attribute " "'score': %e" % (item.id, e))
        if self._context._allow_attributes:
            attrs = self._convert_attributes(ann)
            if attrs:
                elem["attributes"] = attrs

        return elem


class _KeypointsExporter(_InstancesExporter):
    def save_categories(self, dataset):
        label_categories = dataset.categories().get(AnnotationType.label)
        if label_categories is None:
            return
        point_categories = dataset.categories().get(AnnotationType.points)

        for idx, label_cat in enumerate(label_categories.items):
            cat = {
                "id": 1 + idx,
                "name": cast(label_cat.name, str, ""),
                "supercategory": cast(label_cat.parent, str, ""),
                "keypoints": [],
                "skeleton": [],
            }

            if point_categories is not None:
                kp_cat = point_categories.items.get(idx)
                if kp_cat is not None:
                    cat.update(
                        {
                            "keypoints": [str(l) for l in kp_cat.labels],
                            "skeleton": [list(map(int, j)) for j in kp_cat.joints],
                        }
                    )
            self.categories.append(cat)

    def save_annotations(self, item):
        point_annotations = [a for a in item.annotations if a.type == AnnotationType.points]
        if not point_annotations:
            return

        # Create annotations for solitary keypoints annotations
        for points in self.find_solitary_points(item.annotations):
            instance = [points, [], None, points.get_bbox()]
            elem = super().convert_instance(instance, item)
            elem.update(self.convert_points_object(points))
            if not self._stream:
                self.annotations.append(elem)
            else:
                self._temporary_writers.add_anns(elem)

        # Create annotations for complete instance + keypoints annotations
        super().save_annotations(item)

    @classmethod
    def find_solitary_points(cls, annotations):
        annotations = sorted(annotations, key=lambda a: a.group)
        solitary_points = []

        for g_id, group in groupby(annotations, lambda a: a.group):
            if not g_id or g_id and not cls.find_instance_anns(group):
                group = [a for a in group if a.type == AnnotationType.points]
                solitary_points.extend(group)

        return solitary_points

    @staticmethod
    def convert_points_object(ann):
        keypoints = []
        points = ann.points
        visibility = ann.visibility
        for index in range(0, len(points), 2):
            kp = points[index : index + 2]
            state = visibility[index // 2].value
            keypoints.extend([*kp, state])

        num_annotated = len([v for v in visibility if v != Points.Visibility.absent])

        return {
            "keypoints": keypoints,
            "num_keypoints": num_annotated,
        }

    def convert_instance(self, instance, item) -> Optional[Dict]:
        points_ann = find(
            item.annotations,
            lambda x: x.type == AnnotationType.points
            and instance[0].group
            and x.group == instance[0].group,
        )
        if not points_ann:
            return None

        elem = super().convert_instance(instance, item)
        elem.update(self.convert_points_object(points_ann))

        return elem


class _LabelsExporter(_TaskExporter):
    def save_categories(self, dataset):
        label_categories = dataset.categories().get(AnnotationType.label)
        if label_categories is None:
            return

        for idx, cat in enumerate(label_categories.items):
            self.categories.append(
                {
                    "id": 1 + idx,
                    "name": cast(cat.name, str, ""),
                    "supercategory": cast(cat.parent, str, ""),
                }
            )

    def save_annotations(self, item):
        for ann in item.annotations:
            if ann.type != AnnotationType.label:
                continue

            elem = {
                "id": self._get_ann_id(ann),
                "image_id": self._get_image_id(item),
                "category_id": int(ann.label) + 1,
            }
            if "score" in ann.attributes:
                try:
                    elem["score"] = float(ann.attributes["score"])
                except Exception as e:
                    log.warning(
                        "Item '%s': failed to convert attribute " "'score': %e" % (item.id, e)
                    )
            if self._context._allow_attributes:
                attrs = self._convert_attributes(ann)
                if attrs:
                    elem["attributes"] = attrs

            if not self._stream:
                self.annotations.append(elem)
            else:
                self._temporary_writers.add_anns(elem)


class _StuffExporter(_InstancesExporter):
    pass


class _PanopticExporter(_TaskExporter):
    def write(self, path):
        dump_json_file(path, self._data)

        if self._stream:
            self._temporary_writers.merge(path, self._data, None)

    def save_categories(self, dataset):
        label_categories = dataset.categories().get(AnnotationType.label)
        if label_categories is None:
            return

        for idx, cat in enumerate(label_categories.items):
            self.categories.append(
                {
                    "id": 1 + idx,
                    "name": cast(cat.name, str, ""),
                    "supercategory": cast(cat.parent, str, ""),
                    "isthing": 0,  # TODO: can't represent this information yet
                }
            )

    def save_annotations(self, item):
        if not item.media:
            return

        ann_filename = item.id + CocoPath.PANOPTIC_EXT

        segments_info = list()
        masks = []
        next_id = self._min_ann_id
        for ann in item.annotations:
            if ann.type != AnnotationType.mask:
                continue

            if not ann.id:
                ann.id = next_id
                next_id += 1

            segment_info = {}
            segment_info["id"] = ann.id
            segment_info["category_id"] = cast(ann.label, int, -1) + 1
            segment_info["area"] = float(ann.get_area())
            segment_info["bbox"] = [float(p) for p in ann.get_bbox()]
            segment_info["iscrowd"] = cast(ann.attributes.get("is_crowd"), int, 0)
            segments_info.append(segment_info)
            masks.append(ann)

        if not masks:
            return

        pan_format = mask_tools.merge_masks((m.image, m.id) for m in masks)
        save_image(
            osp.join(self._context._segmentation_dir, ann_filename),
            mask_tools.index2bgr(pan_format),
            create_dir=True,
        )

        elem = {
            "image_id": self._get_image_id(item),
            "file_name": ann_filename,
            "segments_info": segments_info,
        }

        if not self._stream:
            self.annotations.append(elem)
        else:
            self._temporary_writers.add_anns(elem)


class CocoExporter(Exporter):
    @staticmethod
    def _split_tasks_string(s):
        return [CocoTask[i.strip()] for i in s.split(",")]

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        kwargs[
            "description"
        ] = """
            Segmentation mask modes ('--segmentation-mode'):|n
            - '{sm.guess.name}': guess the mode for each instance,|n
            |s|suse 'is_crowd' attribute as hint|n
            - '{sm.polygons.name}': save polygons,|n
            |s|smerge and convert masks, prefer polygons|n
            - '{sm.mask.name}': save masks,|n
            |s|smerge and convert polygons, prefer masks|n
            |n
            The '--reindex' option allows to control if the images and
            annotations must be given new indices. It can be useful, when
            you want to preserve the original indices in the produced dataset.
            Consider having this option enabled when converting from other
            formats.|n
            |n
            The '--allow-attributes' parameter enables or disables writing
            the custom annotation attributes to the "attributes" annotation
            field. This field is an extension to the original COCO format.|n
            |n
            The '--merge-images' parameter controls the output directory for
            images. When enabled, the dataset images are saved into a single
            directory, otherwise they are saved in separate directories
            by subsets.
            """.format(
            sm=SegmentationMode
        )
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--segmentation-mode",
            choices=[m.name for m in SegmentationMode],
            default=SegmentationMode.guess.name,
            help="Save mode for instance segmentation (default: %(default)s)",
        )
        parser.add_argument(
            "--crop-covered",
            action="store_true",
            help="Crop covered segments so that background objects' "
            "segmentation was more accurate (default: %(default)s)",
        )
        parser.add_argument(
            "--allow-attributes",
            type=str_to_bool,
            default=True,
            help="Allow export of attributes (default: %(default)s)",
        )
        parser.add_argument(
            "--reindex",
            type=str_to_bool,
            default=True,
            help="Assign new indices to images and annotations, "
            "useful to avoid merge conflicts (default: %(default)s)",
        )
        parser.add_argument(
            "--merge-images",
            type=str_to_bool,
            default=False,
            help="Save all images into a single " "directory (default: %(default)s)",
        )
        parser.add_argument(
            "--tasks",
            type=cls._split_tasks_string,
            help="COCO task filter, comma-separated list of {%s} "
            "(default: all)" % ", ".join(t.name for t in CocoTask),
        )
        return parser

    DEFAULT_IMAGE_EXT = CocoPath.IMAGE_EXT

    _TASK_CONVERTER: Dict[CocoTask, Type[_TaskExporter]] = {
        CocoTask.image_info: _ImageInfoExporter,
        CocoTask.instances: _InstancesExporter,
        CocoTask.person_keypoints: _KeypointsExporter,
        CocoTask.captions: _CaptionsExporter,
        CocoTask.labels: _LabelsExporter,
        CocoTask.panoptic: _PanopticExporter,
        CocoTask.stuff: _StuffExporter,
    }

    def __init__(
        self,
        extractor,
        save_dir,
        tasks=None,
        segmentation_mode=None,
        crop_covered=False,
        allow_attributes=True,
        reindex=False,
        merge_images=False,
        stream: bool = False,
        **kwargs,
    ):
        super().__init__(extractor, save_dir, stream=stream, **kwargs)

        assert tasks is None or isinstance(tasks, (CocoTask, list, str))
        if isinstance(tasks, CocoTask):
            tasks = [tasks]
        elif isinstance(tasks, str):
            tasks = [CocoTask[tasks]]
        elif tasks:
            for i, t in enumerate(tasks):
                if isinstance(t, str):
                    tasks[i] = CocoTask[t]
                else:
                    assert t in CocoTask, t
        else:
            tasks = set()
        self._tasks = tasks

        assert (
            segmentation_mode is None
            or isinstance(segmentation_mode, str)
            or segmentation_mode in SegmentationMode
        )
        if segmentation_mode is None:
            segmentation_mode = SegmentationMode.guess
        if isinstance(segmentation_mode, str):
            segmentation_mode = SegmentationMode[segmentation_mode]
        self._segmentation_mode = segmentation_mode

        self._crop_covered = crop_covered
        self._allow_attributes = allow_attributes
        self._reindex = reindex
        self._merge_images = merge_images

        self._image_ids = {}

        self._patch = None

    def _make_dirs(self):
        self._images_dir = osp.join(self._save_dir, CocoPath.IMAGES_DIR)
        os.makedirs(self._images_dir, exist_ok=True)

        self._ann_dir = osp.join(self._save_dir, CocoPath.ANNOTATIONS_DIR)
        os.makedirs(self._ann_dir, exist_ok=True)

    def _make_segmentation_dir(self, subset_name):
        self._segmentation_dir = osp.join(
            self._save_dir, CocoPath.ANNOTATIONS_DIR, "panoptic_" + subset_name
        )
        os.makedirs(self._segmentation_dir, exist_ok=True)

    def _make_task_converter(self, task: CocoTask, subset: str) -> _TaskExporter:
        if task not in self._TASK_CONVERTER:
            raise NotImplementedError()
        return self._TASK_CONVERTER[task](
            context=self,
            subset=subset,
            task=task,
            ann_dir=self._ann_dir,
            stream=self._stream,
        )

    def _make_task_converters(self, subset: str):
        return {
            task: self._make_task_converter(task, subset)
            for task in (self._tasks or self._TASK_CONVERTER)
        }

    def _get_image_id(self, item):
        image_id = self._image_ids.get(item.id)
        if image_id is None:
            if not self._reindex:
                image_id = cast(item.attributes.get("id"), int, len(self._image_ids) + 1)
            else:
                image_id = len(self._image_ids) + 1
            self._image_ids[item.id] = image_id
        return image_id

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        self._make_dirs()

        if self._save_dataset_meta:
            self._save_meta_file(self._save_dir)

        subsets = self._extractor.subsets()
        pbars = self._ctx.progress_reporter.split(len(subsets))
        for pbar, (subset_name, subset) in zip(pbars, subsets.items()):
            task_converters = self._make_task_converters(subset_name)
            for task_conv in task_converters.values():
                task_conv.save_categories(subset)
            if CocoTask.panoptic in task_converters:
                self._make_segmentation_dir(subset_name)

            for item in pbar.iter(subset, desc=f"Exporting '{subset_name}'"):
                try:
                    if self._save_media:
                        if item.media:
                            self._save_image(
                                item,
                                subdir=osp.join(
                                    self._images_dir, "" if self._merge_images else subset_name
                                ),
                            )
                        else:
                            log.debug("Item '%s' has no image info", item.id)

                    for task_conv in task_converters.values():
                        try:
                            task_conv.save_image_info(item, self._make_image_filename(item))
                            task_conv.save_annotations(item)
                        except Exception as e:
                            self._report_annotation_error(e, item_id=(item.id, item.subset))
                except Exception as e:
                    self._ctx.error_policy.report_item_error(e, item_id=(item.id, item.subset))

            for task, task_conv in task_converters.items():
                ann_file = osp.join(self._ann_dir, "%s_%s.json" % (task.name, subset_name))

                if task_conv.is_empty() and (not self._tasks or self._patch):
                    if task == CocoTask.panoptic:
                        os.rmdir(self._segmentation_dir)
                    if self._patch:
                        if osp.isfile(ann_file):
                            # Remove subsets that became empty
                            os.remove(ann_file)
                    continue

                task_conv.write(ann_file)

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        for subset in patch.updated_subsets:
            conv = cls(dataset.get_subset(subset), save_dir=save_dir, **kwargs)
            conv._patch = patch
            conv.apply()

        conv = cls(dataset, save_dir=save_dir, **kwargs)
        images_dir = osp.join(save_dir, CocoPath.IMAGES_DIR)
        for (item_id, subset), status in patch.updated_items.items():
            if status != ItemStatus.removed:
                item = patch.data.get(item_id, subset)
            else:
                item = DatasetItem(item_id, subset=subset)

            if not (status == ItemStatus.removed or not item.media):
                continue

            # Exporter supports saving in separate dirs and common image dir

            image_path = osp.join(images_dir, conv._make_image_filename(item))
            if osp.isfile(image_path):
                os.unlink(image_path)

            image_path = osp.join(images_dir, subset, conv._make_image_filename(item))
            if osp.isfile(image_path):
                os.unlink(image_path)

    @property
    def can_stream(self) -> bool:
        return True


class CocoInstancesExporter(CocoExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = CocoTask.instances
        super().__init__(*args, **kwargs)


class CocoImageInfoExporter(CocoExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = CocoTask.image_info
        super().__init__(*args, **kwargs)


class CocoPersonKeypointsExporter(CocoExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = CocoTask.person_keypoints
        super().__init__(*args, **kwargs)


class CocoCaptionsExporter(CocoExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = CocoTask.captions
        super().__init__(*args, **kwargs)


class CocoLabelsExporter(CocoExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = CocoTask.labels
        super().__init__(*args, **kwargs)


class CocoPanopticExporter(CocoExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = CocoTask.panoptic
        super().__init__(*args, **kwargs)


class CocoStuffExporter(CocoExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = CocoTask.stuff
        kwargs["segmentation_mode"] = SegmentationMode.mask
        super().__init__(*args, **kwargs)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from typing import Any, Dict, Iterator, Optional, Set

from datumaro.components.annotation import AnnotationType
from datumaro.components.media import MediaType
from datumaro.rust_api import DatumPageMapper as DatumPageMapperImpl

__all__ = ["DatumPageMapper"]


class DatumPageMapper:
    """Construct page maps for items and annotations from the JSON file,
    which are used for the stream importer.

    It also provides __iter__() to produce item and annotation dictionaries
    in stream manner after constructing the page map.
    """

    def __init__(self, path: str) -> None:
        self._path = path
        self._impl = DatumPageMapperImpl(path)

    def __iter__(self) -> Iterator[Dict]:
        for item_key in self.iter_item_ids():
            yield self._impl.get_item_dict(item_key)

    def get_item_dict(self, item_key: str) -> Optional[Dict]:
        try:
            return self._impl.get_item_dict(item_key)
        except Exception as e:
            log.error(e)
            return None

    def __len__(self) -> int:
        return len(self._impl)

    def iter_item_ids(self) -> Iterator[str]:
        for item_id in self._impl.get_img_ids():
            yield item_id

    def __del__(self):
        pass

    @property
    def dm_format_version(self) -> Optional[str]:
        """Parse "dm_format_version" section from the given JSON file using the stream json parser"""
        return self._impl.dm_format_version()

    @property
    def media_type(self) -> Optional[MediaType]:
        """Parse "media_type" section from the given JSON file using the stream json parser"""
        media_type = self._impl.media_type()
        if media_type is not None:
            return MediaType(media_type)
        return None

    @property
    def ann_types(self) -> Optional[Set[AnnotationType]]:
        """Parse "media_type" section from the given JSON file using the stream json parser"""
        ann_types = self._impl.ann_types()
        if ann_types is not None:
            return ann_types
        return None

    @property
    def infos(self) -> Dict[str, Any]:
        """Parse "infos" section from the given JSON file using the stream json parser"""
        return self._impl.infos()

    @property
    def categories(self) -> Dict[str, Any]:
        """Parse "categories" section from the given JSON file using the stream json parser"""
        return self._impl.categories()

    def __reduce__(self):
        return (self.__class__, (self._path,))



# Copyright (C) 2019-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import os
import os.path as osp
import shlex
import weakref
from contextlib import contextmanager
from contextvars import ContextVar
from enum import Enum, auto
from functools import partial
from io import BytesIO, IOBase
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Iterator, Optional, Tuple, Union

import numpy as np

from datumaro.components.crypter import NULL_CRYPTER, Crypter
from datumaro.components.errors import DatumaroError

if TYPE_CHECKING:
    try:
        # Introduced in 1.20
        from numpy.typing import DTypeLike
    except ImportError:
        DTypeLike = Any


class ImageBackend(Enum):
    cv2 = auto()
    PIL = auto()


IMAGE_BACKEND: ContextVar[ImageBackend] = ContextVar("IMAGE_BACKEND")
_image_loading_errors = (FileNotFoundError,)
try:
    import cv2

    IMAGE_BACKEND.set(ImageBackend.cv2)
except ModuleNotFoundError:
    import PIL

    IMAGE_BACKEND.set(ImageBackend.PIL)
    _image_loading_errors = (*_image_loading_errors, PIL.UnidentifiedImageError)

from datumaro.util.image_cache import ImageCache
from datumaro.util.os_util import find_files


class ImageColorChannel(Enum):
    """Image color channel

    - UNCHANGED: Use the original image's channel (default)
    - COLOR_BGR: Use BGR 3 channels (it can ignore the alpha channel or convert the gray scale image)
    - COLOR_RGB: Use RGB 3 channels (it can ignore the alpha channel or convert the gray scale image)
    """

    UNCHANGED = 0
    COLOR_BGR = 1
    COLOR_RGB = 2

    def decode_by_cv2(self, image_bytes: bytes, dtype: DTypeLike = np.uint8) -> np.ndarray:
        """Convert image color channel for OpenCV image (np.ndarray)."""
        image_buffer = np.frombuffer(image_bytes, dtype=dtype)

        if self == ImageColorChannel.UNCHANGED:
            return cv2.imdecode(image_buffer, cv2.IMREAD_UNCHANGED)

        img = cv2.imdecode(image_buffer, cv2.IMREAD_COLOR)

        if self == ImageColorChannel.COLOR_BGR:
            return img

        if self == ImageColorChannel.COLOR_RGB:
            return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        raise ValueError

    def decode_by_pil(self, image_bytes: bytes) -> np.ndarray:
        """Convert image color channel for PIL Image."""
        from PIL import Image

        img = Image.open(BytesIO(image_bytes))

        if self == ImageColorChannel.UNCHANGED:
            return np.asarray(img)

        if self == ImageColorChannel.COLOR_BGR:
            img = np.asarray(img.convert("RGB"))
            return cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

        if self == ImageColorChannel.COLOR_RGB:
            return np.asarray(img.convert("RGB"))

        raise ValueError


IMAGE_COLOR_CHANNEL: ContextVar[ImageColorChannel] = ContextVar(
    "IMAGE_COLOR_CHANNEL", default=ImageColorChannel.UNCHANGED
)


@contextmanager
def decode_image_context(image_backend: ImageBackend, image_color_channel: ImageColorChannel):
    """Change Datumaro image color channel while decoding.

    For model training, it is recommended to use this context manager
    to load images in the BGR 3-channel format. For example,

    .. code-block:: python

        import datumaro as dm
        with decode_image_context(image_backend=ImageBackend.cv2, image_color_channel=ImageColorScale.COLOR):
            item: dm.DatasetItem
            img_data = item.media_as(dm.Image).data
            assert img_data.shape[-1] == 3  # It should be a 3-channel image
    """

    curr_ctx = (IMAGE_BACKEND.get(), IMAGE_COLOR_CHANNEL.get())

    IMAGE_BACKEND.set(image_backend)
    IMAGE_COLOR_CHANNEL.set(image_color_channel)

    yield

    IMAGE_BACKEND.set(curr_ctx[0])
    IMAGE_COLOR_CHANNEL.set(curr_ctx[1])


def load_image(path: str, dtype: DTypeLike = np.uint8, crypter: Crypter = NULL_CRYPTER):
    """
    Reads an image in the HWC Grayscale/BGR(A) [0; 255] format (default dtype is uint8).
    """

    if IMAGE_BACKEND.get() == ImageBackend.cv2:
        # cv2.imread does not support paths that are not representable
        # in the locale encoding on Windows, so we read the image bytes
        # ourselves.

        with open(path, "rb") as f:
            image_bytes = crypter.decrypt(f.read())

        return decode_image(image_bytes, dtype=dtype)
    elif IMAGE_BACKEND.get() == ImageBackend.PIL:
        with open(path, "rb") as f:
            image_bytes = crypter.decrypt(f.read())

        return decode_image(image_bytes, dtype=dtype)

    raise NotImplementedError(IMAGE_BACKEND)


def copyto_image(
    src: Union[str, IOBase], dst: Union[str, IOBase], src_crypter: Crypter, dst_crypter: Crypter
) -> None:
    if src_crypter == dst_crypter and src == dst:
        return

    @contextmanager
    def _open(fp, mode):
        was_file = False
        if not isinstance(fp, IOBase):
            was_file = True
            fp = open(fp, mode)
        yield fp
        if was_file:
            fp.close()

    with _open(src, "rb") as src_fp:
        _bytes = src_crypter.decrypt(src_fp.read())

    with _open(dst, "wb") as dst_fp:
        dst_fp.write(dst_crypter.encrypt(_bytes))


def save_image(
    dst: Union[str, IOBase],
    image: np.ndarray,
    ext: Optional[str] = None,
    create_dir: bool = False,
    dtype: DTypeLike = np.uint8,
    crypter: Crypter = NULL_CRYPTER,
    **kwargs,
) -> None:
    # NOTE: Check destination path for existence
    # OpenCV silently fails if target directory does not exist
    if isinstance(dst, IOBase):
        ext = ext if ext else ".png"
    else:
        dst_dir = osp.dirname(dst)
        if dst_dir:
            if create_dir:
                os.makedirs(dst_dir, exist_ok=True)
            elif not osp.isdir(dst_dir):
                raise FileNotFoundError("Directory does not exist: '%s'" % dst_dir)
        # file extension and actual encoding can be differ
        ext = ext if ext else osp.splitext(dst)[1]

    if not kwargs:
        kwargs = {}

    # NOTE: OpenCV documentation says "If the image format is not supported,
    # the image will be converted to 8-bit unsigned and saved that way".
    # Conversion from np.int32 to np.uint8 is not working properly
    backend = IMAGE_BACKEND.get()
    if dtype == np.int32:
        backend = ImageBackend.PIL
    if backend == ImageBackend.cv2:
        # cv2.imwrite does not support paths that are not representable
        # in the locale encoding on Windows, so we write the image bytes
        # ourselves.

        image_bytes = encode_image(image, ext, dtype=dtype, **kwargs)

        if isinstance(dst, str):
            with open(dst, "wb") as f:
                f.write(crypter.encrypt(image_bytes))
        else:
            dst.write(crypter.encrypt(image_bytes))
    elif backend == ImageBackend.PIL:
        from PIL import Image

        if ext.startswith("."):
            ext = ext[1:]

        if not crypter.is_null_crypter:
            raise DatumaroError("PIL backend should have crypter=NullCrypter.")

        params = {}
        params["quality"] = kwargs.get("jpeg_quality")
        if kwargs.get("jpeg_quality") == 100:
            params["subsampling"] = 0

        image = image.astype(dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # BGR to RGB
        image = Image.fromarray(image)
        image.save(dst, format=ext, **params)
    else:
        raise NotImplementedError()


def encode_image(image: np.ndarray, ext: str, dtype: DTypeLike = np.uint8, **kwargs) -> bytes:
    if not kwargs:
        kwargs = {}

    if IMAGE_BACKEND.get() == ImageBackend.cv2:
        import cv2

        params = []

        if not ext.startswith("."):
            ext = "." + ext

        if ext.upper() in (".JPG", ".JPEG"):
            params = [int(cv2.IMWRITE_JPEG_QUALITY), kwargs.get("jpeg_quality", 75)]

        image = image.astype(dtype)
        success, result = cv2.imencode(ext, image, params=params)
        if not success:
            raise Exception("Failed to encode image to '%s' format" % (ext))
        return result.tobytes()
    elif IMAGE_BACKEND.get() == ImageBackend.PIL:
        from PIL import Image

        if ext.startswith("."):
            ext = ext[1:]

        params = {}
        params["quality"] = kwargs.get("jpeg_quality")
        if kwargs.get("jpeg_quality") == 100:
            params["subsampling"] = 0

        image = image.astype(dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # BGR to RGB
        image = Image.fromarray(image)
        with BytesIO() as buffer:
            image.save(buffer, format=ext, **params)
            return buffer.getvalue()
    else:
        raise NotImplementedError()


def decode_image(image_bytes: bytes, dtype: np.dtype = np.uint8) -> np.ndarray:
    ctx_color_scale = IMAGE_COLOR_CHANNEL.get()

    if np.issubdtype(dtype, np.floating):
        # PIL doesn't support floating point image loading
        # CV doesn't support floating point image with color channel setting (IMREAD_COLOR)
        with decode_image_context(
            image_backend=ImageBackend.cv2, image_color_channel=ImageColorChannel.UNCHANGED
        ):
            image = ctx_color_scale.decode_by_cv2(image_bytes, dtype=dtype)
            image = image[..., ::-1]
        if ctx_color_scale == ImageColorChannel.COLOR_BGR:
            image = image[..., ::-1]
    else:
        if IMAGE_BACKEND.get() == ImageBackend.cv2:
            image = ctx_color_scale.decode_by_cv2(image_bytes)
        elif IMAGE_BACKEND.get() == ImageBackend.PIL:
            image = ctx_color_scale.decode_by_pil(image_bytes)
        else:
            raise NotImplementedError()

    image = image.astype(dtype)

    assert len(image.shape) in {2, 3}
    if len(image.shape) == 3:
        assert image.shape[2] in {3, 4}
    return image


IMAGE_EXTENSIONS = {
    ".jpg",
    ".jpeg",
    ".jpe",
    ".jp2",
    ".png",
    ".bmp",
    ".dib",
    ".tif",
    ".tiff",
    ".tga",
    ".webp",
    ".pfm",
    ".sr",
    ".ras",
    ".exr",
    ".hdr",
    ".pic",
    ".pbm",
    ".pgm",
    ".ppm",
    ".pxm",
    ".pnm",
}


def find_images(
    dirpath: str,
    exts: Union[str, Iterable[str]] = None,
    recursive: bool = False,
    max_depth: Optional[int] = None,
    min_depth: Optional[int] = None,
) -> Iterator[str]:
    yield from find_files(
        dirpath,
        exts=exts or IMAGE_EXTENSIONS,
        recursive=recursive,
        max_depth=max_depth,
        min_depth=min_depth,
    )


def is_image(path: str) -> bool:
    trunk, ext = osp.splitext(osp.basename(path))
    return trunk and ext.lower() in IMAGE_EXTENSIONS and osp.isfile(path)


class lazy_image:
    def __init__(
        self,
        path: str,
        loader: Callable[[str], np.ndarray] = None,
        cache: Union[bool, ImageCache] = True,
        crypter: Crypter = NULL_CRYPTER,
        dtype: Optional[DTypeLike] = None,
    ) -> None:
        """
        Cache:
            - False: do not cache
            - True: use the global cache
            - ImageCache instance: an object to be used as cache
        """

        self._custom_loader = True

        if loader is None:
            loader = partial(load_image, dtype=dtype) if dtype else load_image
            self._custom_loader = False

        self._path = path
        self._loader = loader

        assert isinstance(cache, (ImageCache, bool))
        self._cache = cache
        self._crypter = crypter
        self._dtype = dtype

    def __call__(self) -> np.ndarray:
        image = None
        cache_key = weakref.ref(self)

        cache = self._get_cache()
        if cache is not None:
            image = cache.get(cache_key)

        if image is None:
            image = (
                self._loader(self._path)
                if self._custom_loader
                else self._loader(self._path, crypter=self._crypter)
            )
            if cache is not None:
                cache.push(cache_key, image)
        return image

    def _get_cache(self) -> Optional[ImageCache]:
        if self._cache is True:
            cache = ImageCache.get_instance()
        elif self._cache is False:
            cache = None
        else:
            cache = self._cache
        return cache


ImageMeta = Dict[str, Tuple[int, int]]
"""filename -> height, width"""

DEFAULT_IMAGE_META_FILE_NAME = "images.meta"


def load_image_meta_file(image_meta_path: str) -> ImageMeta:
    """
    Loads image metadata from a file with the following format:

        <image name 1> <height 1> <width 1>

        <image name 2> <height 2> <width 2>

        ...

    Shell-like comments and quoted fields are allowed.

    This can be useful to support datasets in which image dimensions are
    required to interpret annotations.
    """
    assert isinstance(image_meta_path, str)

    if not osp.isfile(image_meta_path):
        raise FileNotFoundError("Can't read image meta file '%s'" % image_meta_path)

    image_meta = {}

    with open(image_meta_path, encoding="utf-8") as f:
        for line in f:
            fields = shlex.split(line, comments=True)
            if not fields:
                continue

            # ignore extra fields, so that the format can be extended later
            image_name, h, w = fields[:3]
            image_meta[image_name] = (int(h), int(w))

    return image_meta


def save_image_meta_file(image_meta: ImageMeta, image_meta_path: str) -> None:
    """
    Saves image_meta to the path specified by image_meta_path in the format
    defined in load_image_meta_file's documentation.
    """

    assert isinstance(image_meta_path, str)

    with open(image_meta_path, "w", encoding="utf-8") as f:
        # Add a comment about file syntax
        print("# <image name> <height> <width>", file=f)
        print("", file=f)

        for image_name, (height, width) in image_meta.items():
            print(shlex.quote(image_name), height, width, file=f)



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.errors import ProjectNotFoundError
from datumaro.components.project import Environment
from datumaro.util.os_util import rmtree
from datumaro.util.scope import on_error_do, scope_add, scoped

from ..util import MultilineFormatter, add_subparser
from ..util.errors import CliException
from ..util.project import (
    generate_next_file_name,
    generate_next_name,
    load_project,
    parse_full_revpath,
)


def build_add_parser(parser_ctor=argparse.ArgumentParser):
    builtins = sorted(Environment().launchers)

    parser = parser_ctor(
        help="Add model to project",
        description="""
        Adds an executable model into a project. A model requires
        a launcher to be executed. Each launcher has its own options, which
        are passed after the '--' separator, pass '-- -h' for more info.
        |n
        List of builtin launchers: {}|n
        |n
        Examples:|n
        - Add an OpenVINO model into a project:|n
        |s|s%(prog)s -l openvino -- -d model.xml -w model.bin -i parse_outs.py
        """.format(
            ", ".join(builtins)
        ),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "-n",
        "--name",
        default=None,
        help="Name of the model to be added (default: generate automatically)",
    )
    parser.add_argument("-l", "--launcher", required=True, help="Model launcher")
    parser.add_argument(
        "--copy", action="store_true", help="Copy model data into project (default: %(default)s)"
    )
    parser.add_argument(
        "--no-check", action="store_true", help="Don't check model loading (default: %(default)s)"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        default=None,
        help="Additional arguments for exporter (pass '-- -h' for help)",
    )
    parser.set_defaults(command=add_command)

    return parser


def get_add_sensitive_args():
    return {
        add_command: ["project_dir", "name", "extra_args"],
    }


@scoped
def add_command(args):
    show_plugin_help = "-h" in args.extra_args or "--help" in args.extra_args

    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if not show_plugin_help and args.project_dir:
            raise

    if project is not None:
        env = project.env
    else:
        env = Environment()

    name = args.name
    if name:
        if name in project.models:
            raise CliException("Model '%s' already exists" % name)
    else:
        name = generate_next_name(list(project.models), "model", sep="-", default=0)

    try:
        launcher = env.launchers[args.launcher]
    except KeyError:
        raise CliException("Launcher '%s' is not found" % args.launcher)

    cli_plugin = getattr(launcher, "cli_plugin", launcher)
    model_args = cli_plugin.parse_cmdline(args.extra_args)

    if args.copy:
        log.info("Copying model data")

        model_dir = project.model_data_dir(name)
        os.makedirs(model_dir, exist_ok=False)
        on_error_do(rmtree, model_dir, ignore_errors=True)

        try:
            cli_plugin.copy_model(model_dir, model_args)
        except (AttributeError, NotImplementedError):
            raise NotImplementedError(
                "Can't copy: copying is not available for '%s' models. " % args.launcher
            )

    project.add_model(name, launcher=args.launcher, options=model_args)
    on_error_do(project.remove_model, name, ignore_errors=True)

    if not args.no_check:
        log.info("Checking the model...")
        project.make_model(name)

    project.save()

    log.info("Model '%s' with launcher '%s' has been added to project", name, args.launcher)

    return 0


def build_remove_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Remove model from project", description="Remove a model from a project"
    )

    parser.add_argument("name", help="Name of the model to be removed")
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=remove_command)

    return parser


def get_remove_sensitive_args():
    return {
        remove_command: [
            "project_dir",
            "name",
        ],
    }


@scoped
def remove_command(args):
    project = scope_add(load_project(args.project_dir))

    project.remove_model(args.name)
    project.save()

    log.info(f"Model {args.name} has been removed from the project.")

    return 0


def build_run_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Launches model inference",
        description="""
        Launches model inference on a dataset.|n
        |n
        Target dataset is specified by a revpath. The full syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        Both forms use the -p/--project as a context for plugins and models.
        When not specified, the current project's working tree is used.|n
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "target", nargs="?", default="project", help="Target dataset revpath (default: %(default)s)"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="Directory to save output (default: auto-generated)",
    )
    parser.add_argument(
        "-m", "--model", dest="model_name", required=True, help="Model to apply to the project"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite output directory if exists"
    )
    parser.set_defaults(command=run_command)

    return parser


def get_run_sensitive_args():
    return {
        run_command: [
            "target",
            "dst_dir",
            "model_name",
            "project_dir",
        ],
    }


@scoped
def run_command(args):
    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name("%s-inference" % args.model_name)
    dst_dir = osp.abspath(dst_dir)

    project = scope_add(load_project(args.project_dir))

    dataset, target_project = parse_full_revpath(args.target, project)
    if target_project:
        scope_add(target_project)

    model = project.make_model(args.model_name)
    inference = dataset.run_model(model)
    inference.save(dst_dir)

    log.info("Inference results have been saved to '%s'" % dst_dir)

    return 0


def build_info_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Print model information in this project",
        description="Print model information in this project",
        formatter_class=MultilineFormatter,
    )

    parser.add_argument("-n", "--name", help="Model name")
    parser.add_argument("-v", "--verbose", action="store_true", help="Show details")
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=info_command)

    return parser


def get_info_sensitive_args():
    return {
        info_command: [
            "name",
            "project_dir",
        ],
    }


@scoped
def info_command(args):
    project = scope_add(load_project(args.project_dir))

    if args.name:
        print(project.models[args.name])
    else:
        for name, conf in project.models.items():
            print(name)
            if args.verbose:
                print(dict(conf))


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor()

    subparsers = parser.add_subparsers()
    add_subparser(subparsers, "add", build_add_parser)
    add_subparser(subparsers, "remove", build_remove_parser)
    add_subparser(subparsers, "run", build_run_parser)
    add_subparser(subparsers, "info", build_info_parser)

    return parser


def get_sensitive_args():
    return {
        **get_add_sensitive_args(),
        **get_remove_sensitive_args(),
        **get_run_sensitive_args(),
        **get_info_sensitive_args(),
    }



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from datumaro.plugins.inference_server_plugin.base import InferenceServerType, ProtocolType
from datumaro.plugins.inference_server_plugin.ovms import OVMSLauncher
from datumaro.plugins.inference_server_plugin.triton import TritonLauncher

__all__ = ["InferenceServerType", "OVMSLauncher", "TritonLauncher", "ProtocolType"]



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT


import logging as log
import os
import os.path as osp
from itertools import chain
from typing import List, Union

from pycocotools import mask as mask_utils

from datumaro.components.annotation import AnnotationType, Ellipse, Polygon
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.media import Image
from datumaro.util import annotation_util as anno_tools
from datumaro.util import dump_json_file, mask_tools


class SegmentAnythingExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".jpg"

    _polygon_types = {AnnotationType.polygon, AnnotationType.ellipse}
    _allowed_types = {
        AnnotationType.bbox,
        AnnotationType.polygon,
        AnnotationType.mask,
        AnnotationType.ellipse,
    }

    def __init__(
        self,
        extractor,
        save_dir,
        **kwargs,
    ):
        super().__init__(extractor, save_dir, **kwargs)

    @staticmethod
    def find_instance_anns(annotations):
        return [a for a in annotations if a.type in SegmentAnythingExporter._allowed_types]

    @classmethod
    def find_instances(cls, annotations):
        return anno_tools.find_instances(cls.find_instance_anns(annotations))

    def get_annotation_info(self, group, img_width, img_height):
        boxes = [a for a in group if a.type == AnnotationType.bbox]
        polygons: List[Union[Polygon, Ellipse]] = [
            a for a in group if a.type in self._polygon_types
        ]
        masks = [a for a in group if a.type == AnnotationType.mask]

        anns = boxes + polygons + masks
        leader = anno_tools.find_group_leader(anns)
        if len(boxes) > 0:
            bbox = anno_tools.max_bbox(boxes)
        else:
            bbox = anno_tools.max_bbox(anns)
        polygons = [p.as_polygon() for p in polygons]

        mask = None
        if polygons:
            mask = mask_tools.rles_to_mask(polygons, img_width, img_height)
        if masks:
            masks = (m.image for m in masks)
            if mask is not None:
                masks = chain(masks, [mask])
            mask = mask_tools.merge_masks(masks)
        if mask is None:
            return None
        mask = mask_tools.mask_to_rle(mask)

        segmentation = {
            "counts": list(int(c) for c in mask["counts"]),
            "size": list(int(c) for c in mask["size"]),
        }
        rles = mask_utils.frPyObjects(segmentation, img_height, img_width)
        if isinstance(rles["counts"], bytes):
            rles["counts"] = rles["counts"].decode()
        area = mask_utils.area(rles)

        annotation_data = {
            "id": leader.group,
            "segmentation": rles,
            "bbox": bbox,
            "area": area,
            "predicted_iou": max(ann.attributes.get("predicted_iou", 0.0) for ann in anns),
            "stability_score": max(ann.attributes.get("stability_score", 0.0) for ann in anns),
            "crop_box": anno_tools.max_bbox([ann.attributes.get("crop_box", []) for ann in anns]),
            "point_coords": list(
                set(
                    tuple(point_coord)
                    for ann in anns
                    for point_coord in ann.attributes.get("point_coords", [[]])
                )
            ),
        }
        return annotation_data

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        os.makedirs(self._save_dir, exist_ok=True)

        subsets = self._extractor.subsets()
        pbars = self._ctx.progress_reporter.split(len(subsets))

        max_image_id = 1
        for pbar, (subset_name, subset) in zip(pbars, subsets.items()):
            for item in pbar.iter(subset, desc=f"Exporting {subset_name}"):
                try:
                    # make sure file_name is flat
                    file_name = self._make_image_filename(item).replace("/", "__")
                    try:
                        image_id = int(item.attributes.get("id", max_image_id))
                    except ValueError:
                        image_id = max_image_id
                    max_image_id += 1

                    if not item.media or not item.media.size:
                        log.warning(
                            f"Item '{item.id}': skipping writing instances since no image info available"
                        )
                        continue

                    height, width = item.media.size
                    json_data = {
                        "image": {
                            "image_id": image_id,
                            "file_name": file_name,
                            "height": height,
                            "width": width,
                        },
                        "annotations": [],
                    }

                    instances = self.find_instances(item.annotations)
                    annotations = [self.get_annotation_info(i, width, height) for i in instances]
                    annotations = [i for i in annotations if i is not None]
                    if not annotations:
                        log.warning(
                            f"Item '{item.id}': skipping writing instances since no annotation available"
                        )
                        continue
                    json_data["annotations"] = annotations

                    dump_json_file(
                        os.path.join(self._save_dir, osp.splitext(file_name)[0] + ".json"),
                        json_data,
                    )

                    if self._save_media:
                        self._save_image(
                            item,
                            path=osp.abspath(osp.join(self._save_dir, file_name)),
                        )

                except Exception as e:
                    self._ctx.error_policy.report_item_error(e, item_id=(item.id, item.subset))



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
import os.path as osp
from collections import OrderedDict, defaultdict
from itertools import chain

# Disable B406: import_xml_sax - the library is used for writing
from xml.sax.saxutils import XMLGenerator  # nosec

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.dataset_item_storage import ItemStatus
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.media import Image
from datumaro.util import cast, mask_tools, pairs

from .format import CvatPath


class XmlAnnotationWriter:
    VERSION = "1.1"

    def __init__(self, f):
        self.xmlgen = XMLGenerator(f, "utf-8")
        self._level = 0

    def _indent(self, newline=True):
        if newline:
            self.xmlgen.ignorableWhitespace("\n")
        self.xmlgen.ignorableWhitespace("  " * self._level)

    def _add_version(self):
        self._indent()
        self.xmlgen.startElement("version", {})
        self.xmlgen.characters(self.VERSION)
        self.xmlgen.endElement("version")

    def open_root(self):
        self.xmlgen.startDocument()
        self.xmlgen.startElement("annotations", {})
        self._level += 1
        self._add_version()

    def _add_meta(self, meta):
        self._level += 1
        for k, v in meta.items():
            if isinstance(v, OrderedDict):
                self._indent()
                self.xmlgen.startElement(k, {})
                self._add_meta(v)
                self._indent()
                self.xmlgen.endElement(k)
            elif isinstance(v, list):
                self._indent()
                self.xmlgen.startElement(k, {})
                for tup in v:
                    self._add_meta(OrderedDict([tup]))
                self._indent()
                self.xmlgen.endElement(k)
            else:
                self._indent()
                self.xmlgen.startElement(k, {})
                self.xmlgen.characters(v)
                self.xmlgen.endElement(k)
        self._level -= 1

    def write_meta(self, meta):
        self._indent()
        self.xmlgen.startElement("meta", {})
        self._add_meta(meta)
        self._indent()
        self.xmlgen.endElement("meta")

    def open_track(self, track):
        self._indent()
        self.xmlgen.startElement("track", track)
        self._level += 1

    def open_image(self, image):
        self._indent()
        self.xmlgen.startElement("image", image)
        self._level += 1

    def open_box(self, box):
        self._indent()
        self.xmlgen.startElement("box", box)
        self._level += 1

    def open_polygon(self, polygon):
        self._indent()
        self.xmlgen.startElement("polygon", polygon)
        self._level += 1

    def open_polyline(self, polyline):
        self._indent()
        self.xmlgen.startElement("polyline", polyline)
        self._level += 1

    def open_points(self, points):
        self._indent()
        self.xmlgen.startElement("points", points)
        self._level += 1

    def open_mask(self, mask):
        self._indent()
        self.xmlgen.startElement("mask", mask)
        self._level += 1

    def open_tag(self, tag):
        self._indent()
        self.xmlgen.startElement("tag", tag)
        self._level += 1

    def add_attribute(self, attribute):
        self._indent()
        self.xmlgen.startElement("attribute", {"name": attribute["name"]})
        self.xmlgen.characters(attribute["value"])
        self.xmlgen.endElement("attribute")

    def _close_element(self, element):
        self._level -= 1
        self._indent()
        self.xmlgen.endElement(element)

    def close_box(self):
        self._close_element("box")

    def close_polygon(self):
        self._close_element("polygon")

    def close_polyline(self):
        self._close_element("polyline")

    def close_points(self):
        self._close_element("points")

    def close_mask(self):
        self._close_element("mask")

    def close_tag(self):
        self._close_element("tag")

    def close_image(self):
        self._close_element("image")

    def close_track(self):
        self._close_element("track")

    def close_root(self):
        self._close_element("annotations")
        self.xmlgen.endDocument()


class _SubsetWriter:
    def __init__(self, file, name, extractor, context):
        self._writer = XmlAnnotationWriter(file)
        self._name = name
        self._extractor = extractor
        self._context = context
        self._item_count = 0

    def is_empty(self):
        return self._item_count == 0

    def write(self):
        self._writer.open_root()
        self._write_meta()

        if self._context._use_track:
            tracks = self._get_tracks()
            for track in tracks:
                self._write_track(track)
        else:
            for index, item in enumerate(self._extractor):
                self._write_item(item, index)

        self._writer.close_root()

    def _write_track(self, track):
        track_id = track["track_id"]
        label_name = self._get_label(track["label"]).name
        annotations = track["annotations"]

        track_info = {"id": str(track_id), "label": label_name}

        self._writer.open_track(track_info)
        for ann in annotations:
            if ann.type in CvatPath.SUPPORTED_EXPORT_SHAPES:
                self._write_shape(ann, write_label_info=False, write_frame=True)
        self._writer.close_track()

    def _get_tracks(self):
        track_infos = defaultdict(lambda: defaultdict(list))
        for item in self._extractor:
            if "frame" not in item.attributes:
                continue
            frame = item.attributes["frame"]
            for annotation in item.annotations:
                if "track_id" not in annotation.attributes:
                    continue
                track_id = int(annotation.attributes["track_id"])
                track_infos[track_id][frame].append(annotation)

        tracks = []

        for tid in sorted(list(track_infos.keys())):
            annotations = []
            for frame in sorted(list(track_infos[tid].keys())):
                for annotation in track_infos[tid][frame]:
                    annotation.attributes["frame"] = frame
                    del annotation.attributes["track_id"]
                    annotations.append(annotation)

            if len(annotations) == 0:
                continue

            valid_label = True
            for annotation in annotations[1:]:
                if annotation.label != annotations[0].label:
                    valid_label = False

            if not valid_label:
                continue

            tracks.append(
                {"track_id": tid, "label": annotations[0].label, "annotations": annotations}
            )

        return tracks

    def _write_item(self, item, index):
        if not self._context._reindex:
            index = cast(item.attributes.get("frame"), int, index)
        image_info = OrderedDict(
            [
                ("id", str(index)),
            ]
        )
        filename = self._context._make_image_filename(item)
        image_info["name"] = filename
        if item.media:
            size = item.media.size
            if size:
                h, w = size
                image_info["width"] = str(w)
                image_info["height"] = str(h)

            if self._context._save_media:
                self._context._save_image(item, osp.join(self._context._images_dir, filename))
        else:
            log.debug("Item '%s' has no image info", item.id)
        self._writer.open_image(image_info)

        for ann in item.annotations:
            if ann.type in CvatPath.SUPPORTED_EXPORT_SHAPES:
                self._write_shape(ann, item)
            elif ann.type == AnnotationType.label:
                self._write_tag(ann, item)
            else:
                continue

        self._writer.close_image()

        self._item_count += 1

    def _write_meta(self):
        label_cat = self._extractor.categories().get(AnnotationType.label, LabelCategories())

        task_items = [
            ("id", ""),
            ("name", self._name),
            ("size", str(len(self._extractor))),
            ("mode", "annotation"),
            ("overlap", ""),
            ("start_frame", "0"),
            ("stop_frame", str(len(self._extractor))),
            ("frame_filter", ""),
            ("z_order", "True"),
            (
                "labels",
                [
                    (
                        "label",
                        OrderedDict(
                            [
                                ("name", label.name),
                                (
                                    "attributes",
                                    [
                                        (
                                            "attribute",
                                            OrderedDict(
                                                [
                                                    ("name", attr),
                                                    ("mutable", "True"),
                                                    ("input_type", "text"),
                                                    ("default_value", ""),
                                                    ("values", ""),
                                                ]
                                            ),
                                        )
                                        for attr in self._get_label_attrs(label)
                                    ],
                                ),
                            ]
                        ),
                    )
                    for label in label_cat.items
                ],
            ),
        ]

        if self._context._original_size is not None:
            task_items.append(
                (
                    "original_size",
                    OrderedDict(
                        [
                            ("width", str(self._context._original_size[0])),
                            ("height", str(self._context._original_size[1])),
                        ]
                    ),
                )
            )

        meta = OrderedDict(
            [
                ("task", OrderedDict(task_items)),
            ]
        )
        self._writer.write_meta(meta)

    def _get_label(self, label_id):
        if label_id is None:
            return ""
        label_cat = self._extractor.categories().get(AnnotationType.label, LabelCategories())
        return label_cat.items[label_id]

    def _get_label_attrs(self, label):
        label_cat = self._extractor.categories().get(AnnotationType.label, LabelCategories())
        if isinstance(label, int):
            label = label_cat[label]
        return set(chain(label.attributes, label_cat.attributes)) - self._context._builtin_attrs

    def _write_shape(self, shape, item=None, write_label_info=True, write_frame=False):
        item_id = "None" if item is None else item.id

        if write_label_info and shape.label is None:
            log.warning("Item %s: skipping a %s with no label", item_id, shape.type.name)
            return

        if write_frame and "frame" not in shape.attributes:
            log.warning("Skipping a %s with no frame", shape.type.name)
            return

        if write_label_info:
            label_name = self._get_label(shape.label).name
            shape_data = OrderedDict([("label", label_name)])
        else:
            shape_data = OrderedDict()

        if write_frame:
            shape_data.update(OrderedDict([("frame", str(int(shape.attributes.get("frame", 0))))]))
        shape_data.update(
            OrderedDict([("occluded", str(int(shape.attributes.get("occluded", False))))])
        )
        shape_data.update(
            OrderedDict([("outside", str(int(shape.attributes.get("outside", False))))])
        )
        shape_data.update(
            OrderedDict([("keyframe", str(int(shape.attributes.get("keyframe", False))))])
        )

        if shape.type == AnnotationType.bbox:
            shape_data.update(
                OrderedDict(
                    [
                        ("xtl", "{:.2f}".format(shape.points[0])),
                        ("ytl", "{:.2f}".format(shape.points[1])),
                        ("xbr", "{:.2f}".format(shape.points[2])),
                        ("ybr", "{:.2f}".format(shape.points[3])),
                    ]
                )
            )
        elif shape.type == AnnotationType.mask:
            # From the manual test for the dataset exported from the CVAT 2.5,
            # the RLE encoding in the dataset has (W, H) binary 2D np.ndarray, not (H, W)
            # Therefore, we need to tranpose it to make its shape as (H, W).
            mask = shape.image.transpose()
            rle_uncompressed = mask_tools.mask_to_rle(mask)
            width, height = mask.shape
            shape_data.update(
                OrderedDict(
                    rle=", ".join([str(c) for c in rle_uncompressed["counts"]]),
                    left=str(0),
                    top=str(0),
                    width=str(width),
                    height=str(height),
                )
            )
        else:
            shape_data.update(
                OrderedDict(
                    [
                        (
                            "points",
                            ";".join(
                                (
                                    ",".join(("{:.2f}".format(x), "{:.2f}".format(y)))
                                    for x, y in pairs(shape.points)
                                )
                            ),
                        ),
                    ]
                )
            )

        shape_data["z_order"] = str(int(shape.z_order))
        if shape.group:
            shape_data["group_id"] = str(shape.group)

        if shape.type == AnnotationType.bbox:
            self._writer.open_box(shape_data)
        elif shape.type == AnnotationType.polygon:
            self._writer.open_polygon(shape_data)
        elif shape.type == AnnotationType.polyline:
            self._writer.open_polyline(shape_data)
        elif shape.type == AnnotationType.points:
            self._writer.open_points(shape_data)
        elif shape.type == AnnotationType.mask:
            self._writer.open_mask(shape_data)
        else:
            raise NotImplementedError("unknown shape type")

        if write_label_info:
            for attr_name, attr_value in shape.attributes.items():
                if attr_name in self._context._builtin_attrs:
                    continue
                if isinstance(attr_value, bool):
                    attr_value = "true" if attr_value else "false"
                if self._context._allow_undeclared_attrs or attr_name in self._get_label_attrs(
                    shape.label
                ):
                    self._writer.add_attribute(
                        OrderedDict(
                            [
                                ("name", str(attr_name)),
                                ("value", str(attr_value)),
                            ]
                        )
                    )
                else:
                    log.warning(
                        "Item %s: skipping undeclared "
                        "attribute '%s' for label '%s' "
                        "(allow with --allow-undeclared-attrs option)",
                        item_id,
                        attr_name,
                        label_name,
                    )

        if shape.type == AnnotationType.bbox:
            self._writer.close_box()
        elif shape.type == AnnotationType.polygon:
            self._writer.close_polygon()
        elif shape.type == AnnotationType.polyline:
            self._writer.close_polyline()
        elif shape.type == AnnotationType.points:
            self._writer.close_points()
        elif shape.type == AnnotationType.mask:
            self._writer.close_mask()
        else:
            raise NotImplementedError("unknown shape type")

    def _write_tag(self, label, item):
        if label.label is None:
            log.warning("Item %s: skipping a %s with no label", item.id, label.type.name)
            return

        label_name = self._get_label(label.label).name
        tag_data = OrderedDict(
            [
                ("label", label_name),
            ]
        )
        if label.group:
            tag_data["group_id"] = str(label.group)
        self._writer.open_tag(tag_data)

        for attr_name, attr_value in label.attributes.items():
            if attr_name in self._context._builtin_attrs:
                continue
            if isinstance(attr_value, bool):
                attr_value = "true" if attr_value else "false"
            if self._context._allow_undeclared_attrs or attr_name in self._get_label_attrs(
                label.label
            ):
                self._writer.add_attribute(
                    OrderedDict(
                        [
                            ("name", str(attr_name)),
                            ("value", str(attr_value)),
                        ]
                    )
                )
            else:
                log.warning(
                    "Item %s: skipping undeclared "
                    "attribute '%s' for label '%s' "
                    "(allow with --allow-undeclared-attrs option)",
                    item.id,
                    attr_name,
                    label_name,
                )

        self._writer.close_tag()


class CvatExporter(Exporter):
    DEFAULT_IMAGE_EXT = CvatPath.IMAGE_EXT

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--reindex",
            action="store_true",
            help="Assign new indices to frames (default: %(default)s)",
        )
        parser.add_argument(
            "--allow-undeclared-attrs",
            action="store_true",
            help="Write annotation attributes even if they are not present in "
            "the input dataset metainfo (default: %(default)s)",
        )
        return parser

    def __init__(
        self,
        extractor,
        save_dir,
        reindex=False,
        allow_undeclared_attrs=False,
        use_track=False,
        **kwargs,
    ):
        super().__init__(extractor, save_dir, **kwargs)

        self._reindex = reindex
        self._builtin_attrs = CvatPath.BUILTIN_ATTRS
        self._allow_undeclared_attrs = allow_undeclared_attrs
        self._use_track = use_track
        if use_track:
            self._original_size = self._get_original_size(extractor)
        else:
            self._original_size = None

    def _get_original_size(self, extractor):
        item_hw = None
        for item in extractor:
            if item_hw is None:
                item_hw = item.media.size
            elif item_hw != item.media.size:
                return None
        return (item_hw[1], item_hw[0])

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        self._images_dir = osp.join(self._save_dir, CvatPath.IMAGES_DIR)
        os.makedirs(self._images_dir, exist_ok=True)

        for subset_name, subset in self._extractor.subsets().items():
            ann_path = osp.join(self._save_dir, "%s.xml" % subset_name)
            with open(ann_path, "w", encoding="utf-8") as f:
                writer = _SubsetWriter(f, subset_name, subset, self)
                writer.write()

            if self._patch and subset_name in self._patch.updated_subsets and writer.is_empty():
                os.remove(ann_path)

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        for subset in patch.updated_subsets:
            conv = cls(dataset.get_subset(subset), save_dir=save_dir, **kwargs)
            conv._patch = patch
            conv.apply()

        conv = cls(dataset, save_dir=save_dir, **kwargs)
        # Find images that needs to be removed
        # images from different subsets are stored in the common directory
        # Avoid situations like:
        # (a, test): added
        # (a, train): removed
        # where the second line removes images from the first.
        ids_to_remove = {}
        for (item_id, subset), status in patch.updated_items.items():
            if status != ItemStatus.removed:
                item = patch.data.get(item_id, subset)
            else:
                item = DatasetItem(item_id, subset=subset)

            if not (status == ItemStatus.removed or not item.media):
                ids_to_remove[item_id] = (item, False)
            else:
                ids_to_remove.setdefault(item_id, (item, True))

        for item, to_remove in ids_to_remove.values():
            if not to_remove:
                continue

            image_path = osp.join(save_dir, CvatPath.IMAGES_DIR, conv._make_image_filename(item))
            if osp.isfile(image_path):
                os.unlink(image_path)



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from collections import defaultdict
from typing import TYPE_CHECKING, Optional, Union

from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import IDataset
from datumaro.components.transformer import Transform
from datumaro.util import parse_str_enum_value

from .algorithm.algorithm import Algorithm, SamplingMethod

if TYPE_CHECKING:
    import pandas as pd
else:
    from datumaro.util.import_util import lazy_import

    pd = lazy_import("pandas")


class RelevancySampler(Transform, CliPlugin):
    r"""
    Sampler that analyzes model inference results on the dataset |n
    and picks the best sample for training.|n
    |n
    Creates a dataset from the `-k/--count` hardest items for a model.
    The whole dataset or a single subset will be split into the `sampled`
    and `unsampled` subsets based on the model confidence.
    The dataset **must** contain model confidence
    values in the `scores` attributes of annotations.|n
    |n
    There are five methods of sampling (the `-m/--method` option):|n
    |s|s- `topk` - Return the k items with the highest uncertainty data|n
    |s|s- `lowk` - Return the k items with the lowest uncertainty data|n
    |s|s- `randk` - Return random k items|n
    |s|s- `mixk` - Return a half using topk, and the other half using lowk method|n
    |s|s- `randtopk` - Select 3*k items randomly, and return the topk among them|n
    |n
    Notes:|n
    |s|s- Each image's inference result must contain the probability for|n
    |s|s|s|sall classes.|n
    |s|s- Requesting a sample larger than the number of all images will|n
    |s|s|s|sreturn all images.|n
    |n
    Example: select the most relevant data subset of 20 images |n
    |s|sbased on model certainty, put the result into 'sample' subset |n
    |s|sand put all the rest into 'unsampled' subset, use 'train' subset |n
    |s|sas input. |n

    .. code-block::

    |s|s%(prog)s \ |n
    |s|s|s|s--algorithm entropy \ |n
    |s|s|s|s--subset_name train \ |n
    |s|s|s|s--sample_name sample \ |n
    |s|s|s|s--unsampled_name unsampled \ |n
    |s|s|s|s--sampling_method topk -k 20
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-k", "--count", type=int, required=True, help="Number of items to sample"
        )
        parser.add_argument(
            "-a",
            "--algorithm",
            default=Algorithm.entropy.name,
            choices=[t.name for t in Algorithm],
            help="Sampling algorithm (one of {}; default: %(default)s)".format(
                ", ".join(t.name for t in Algorithm)
            ),
        )
        parser.add_argument(
            "-i",
            "--input_subset",
            default=None,
            help="Subset name to select sample from (default: %(default)s)",
        )
        parser.add_argument(
            "-o",
            "--sampled_subset",
            default="sample",
            help="Subset name to put sampled data to (default: %(default)s)",
        )
        parser.add_argument(
            "-u",
            "--unsampled_subset",
            default="unsampled",
            help="Subset name to put the rest data to (default: %(default)s)",
        )
        parser.add_argument(
            "-m",
            "--sampling_method",
            default=SamplingMethod.topk.name,
            choices=[t.name for t in SamplingMethod],
            help="Sampling method (one of {}; default: %(default)s)".format(
                ", ".join(t.name for t in SamplingMethod)
            ),
        )
        parser.add_argument("-d", "--output_file", help="A .csv file path to dump sampling results")
        return parser

    def __init__(
        self,
        extractor: IDataset,
        count: int,
        *,
        algorithm: Union[str, Algorithm],
        sampling_method: Union[str, SamplingMethod],
        input_subset: Optional[str] = None,
        sampled_subset: str = "sample",
        unsampled_subset: str = "unsampled",
        output_file: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        extractor
        algorithm
            Specifying the algorithm to calculate the uncertainty
            for sample selection. default: 'entropy'
        subset_name
            The name of the subset to which you want to select a sample.
        sample_name
            Subset name of the selected sample, default: 'sample'
        sampling_method
            Method of sampling, 'topk' or 'lowk' or 'randk'
        count
            Number of samples extracted
        output_file
            A path to .csv file for sampling results
        """
        super().__init__(extractor)

        self.input_subset = input_subset
        self.sampled_subset = sampled_subset
        self.unsampled_subset = unsampled_subset
        self.algorithm = parse_str_enum_value(algorithm, Algorithm).name
        self.sampling_method = parse_str_enum_value(sampling_method, SamplingMethod).name
        self.count = count
        self.output_file = output_file

        # Use the --output_file option to save the sample list as a csv file
        if output_file and not output_file.endswith(".csv"):
            raise ValueError("The output file must have the '.csv' extension")

    @staticmethod
    def _load_inference_from_subset(extractor, subset_name):
        # 1. Get Dataset from subset name
        if subset_name in extractor.subsets():
            subset = extractor.get_subset(subset_name)
        else:
            raise Exception(f"Unknown subset '{subset_name}'")

        data_df = defaultdict(list)
        infer_df = defaultdict(list)

        # 2. Fill the data_df and infer_df to fit the sampler algorithm
        # input format.
        for item in subset:
            data_df["ImageID"].append(item.id)

            if not item.media or item.media.size is None:
                raise Exception(f"Item {item.id} does not have image info")

            width, height = item.media.size
            data_df["Width"].append(width)
            data_df["Height"].append(height)
            data_df["ImagePath"].append(getattr(item.media, "path", None))

            if not item.annotations:
                raise Exception(f"Item {item.id} does not have annotations")

            for annotation in item.annotations:
                if "scores" not in annotation.attributes:
                    raise Exception(
                        f"Item {item.id} - an annotation " "does not have 'scores' attribute"
                    )
                probs = annotation.attributes["scores"]

                infer_df["ImageID"].append(item.id)

                for prob_idx, prob in enumerate(probs):
                    infer_df[f"ClassProbability{prob_idx+1}"].append(prob)

        data_df = pd.DataFrame(data_df)
        infer_df = pd.DataFrame(infer_df)

        return data_df, infer_df

    @staticmethod
    def _calculate_uncertainty(algorithm, data, inference):
        # Checking and creating algorithms
        if algorithm == Algorithm.entropy.name:
            from .algorithm.entropy import SampleEntropy

            # Data delivery, uncertainty score calculations also proceed.
            sampler = SampleEntropy(data, inference)
        else:
            raise Exception(
                f"Unknown algorithm '{algorithm}', available "
                f"algorithms: {[a.name for a in Algorithm]}"
            )
        return sampler

    def _get_sample_subset(self, image):
        if image.subset == self.input_subset:
            # 1. Returns the sample subset if the id belongs to samples.
            if image.id in self.sample_id:
                return self.sampled_subset
            else:
                return self.unsampled_subset
        else:
            # 2. Returns the existing subset name if it is not a sample
            return image.subset

    def __iter__(self):
        # Import data into a subset name and convert it
        # to a format that will be used in the sampler algorithm with
        # the inference result.
        data_df, infer_df = self._load_inference_from_subset(self._extractor, self.input_subset)

        sampler = self._calculate_uncertainty(self.algorithm, data_df, infer_df)
        self.result = sampler.get_sample(method=self.sampling_method, k=self.count)

        if self.output_file is not None:
            self.result.to_csv(self.output_file, index=False)

        self.sample_id = self.result["ImageID"].to_list()

        # Transform properties for each data
        for item in self._extractor:
            # After checking whether each item belongs to a sample,
            # rename the subset
            yield self.wrap_item(item, subset=self._get_sample_subset(item))



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log

from datumaro.cli.util.errors import CliException, WrongRevpathError
from datumaro.components.errors import ConflictingCategoriesError, ProjectNotFoundError
from datumaro.components.operations import compute_ann_statistics, compute_image_statistics
from datumaro.util import dump_json_file, str_to_bool
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.project import generate_next_file_name, load_project, parse_full_revpath

__all__ = [
    "build_parser",
    "get_sensitive_args",
]


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Get project statistics",
        description="""
        Outputs various project statistics like image mean and std (RGB),
        annotations count etc.|n
        |n
        Target dataset is specified by a revpath. The full syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        Both forms use the -p/--project as a context for plugins. It can be
        useful for dataset paths in targets. When not specified, the current
        project's working tree is used.|n
        |n
        Examples:|n
        - Compute project statistics:|n
        |s|s%(prog)s
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "target", default="project", nargs="?", help="Target dataset revpath (default: project)"
    )
    parser.add_argument("-s", "--subset", help="Compute stats only for a specific subset")
    parser.add_argument(
        "--image-stats",
        type=str_to_bool,
        default=True,
        help="Compute image mean and std (RGB) (default: %(default)s)",
    )
    parser.add_argument(
        "--ann-stats",
        type=str_to_bool,
        default=True,
        help="Compute annotation statistics (default: %(default)s)",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=stats_command)

    return parser


def get_sensitive_args():
    return {
        stats_command: ["project_dir", "target"],
    }


@scoped
def stats_command(args):
    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if args.project_dir:
            raise

    try:
        dataset, target_project = parse_full_revpath(args.target, project)
    except WrongRevpathError as e:
        for p in e.problems:
            if isinstance(p, ConflictingCategoriesError):
                src_names = [src for src in project.working_tree.sources]
                raise CliException(
                    "There are more than two sources with heterogeneous categories in the project. "
                    "This prevents computing the statistics of the merged one. "
                    f"Please specify one of the sources in the project ({src_names}), "
                    f"such as `datum stats {src_names[0]}`"
                ) from e

        raise e

    if target_project:
        scope_add(target_project)

    if args.subset:
        dataset = dataset.get_subset(args.subset)

    stats = {}
    if args.image_stats:
        stats.update(compute_image_statistics(dataset))
    if args.ann_stats:
        stats.update(compute_ann_statistics(dataset))

    dst_file = generate_next_file_name("statistics", ext=".json")
    log.info("Writing project statistics to '%s'" % dst_file)
    dump_json_file(dst_file, stats, indent=True)



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log

from datumaro.util.scope import scope_add, scoped

from ....util.errors import CliException
from ....util.project import load_project

__all__ = [
    "build_parser",
    "get_sensitive_args",
]


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Remove source from project", description="Remove a source from a project"
    )

    parser.add_argument("names", nargs="+", help="Names of the sources to be removed")
    parser.add_argument(
        "--force", action="store_true", help="Do not fail and stop on errors during removal"
    )
    parser.add_argument(
        "--keep-data",
        action="store_true",
        help="Do not remove source data from the working directory, remove "
        "only project metainfo.",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=remove_command)

    return parser


def get_sensitive_args():
    return {
        remove_command: [
            "project_dir",
            "names",
        ],
    }


@scoped
def remove_command(args):
    project = scope_add(load_project(args.project_dir))

    if not args.names:
        raise CliException("Expected source name")

    for name in args.names:
        project.remove_source(name, force=args.force, keep_data=args.keep_data)
    project.working_tree.save()

    log.info("Sources '%s' have been removed from the project" % ", ".join(args.names))

    return 0



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from collections import defaultdict
from typing import Dict, Iterator, List, Optional, Sequence, TypeVar

from datumaro.components.contexts.importer import _ImportFail
from datumaro.components.dataset_base import (
    CategoriesInfo,
    DatasetBase,
    DatasetInfo,
    DatasetItem,
    SubsetBase,
)

T = TypeVar("T")


def check_identicalness(seq: Sequence[T], raise_error_on_empty: bool = True) -> Optional[T]:
    if len(seq) == 0 and raise_error_on_empty:
        raise _ImportFail("It should not be empty.")
    elif len(seq) == 0 and not raise_error_on_empty:
        return None

    if seq.count(seq[0]) != len(seq):
        raise _ImportFail("All items in the sequence should be identical.")

    return seq[0]


class ExtractorMerger(DatasetBase):
    """A simple class to merge single-subset extractors."""

    def __init__(
        self,
        sources: Sequence[SubsetBase],
    ):
        if len(sources) == 0:
            raise _ImportFail("It should not be empty.")

        self._infos = check_identicalness([s.infos() for s in sources])
        self._categories = check_identicalness([s.categories() for s in sources])
        self._media_type = check_identicalness([s.media_type() for s in sources])

        ann_types = set()
        for source in sources:
            ann_types.union(source.ann_types())
        self._ann_types = ann_types

        self._is_stream = check_identicalness([s.is_stream for s in sources])

        self._subsets: Dict[str, List[SubsetBase]] = defaultdict(list)
        for source in sources:
            self._subsets[source.subset] += [source]

    def infos(self) -> DatasetInfo:
        return self._infos

    def categories(self) -> CategoriesInfo:
        return self._categories

    def __iter__(self) -> Iterator[DatasetItem]:
        for sources in self._subsets.values():
            for source in sources:
                yield from source

    def __len__(self) -> int:
        return sum(len(source) for sources in self._subsets.values() for source in sources)

    def get(self, id: str, subset: Optional[str] = None) -> Optional[DatasetItem]:
        if subset is not None and (sources := self._subsets.get(subset, [])):
            for source in sources:
                if item := source.get(id, subset):
                    return item

        for sources in self._subsets.values():
            for source in sources:
                if item := source.get(id=id, subset=source.subset):
                    return item

        return None

    @property
    def is_stream(self) -> bool:
        return self._is_stream



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from typing import Optional, Union

from cryptography.fernet import Fernet, InvalidToken

from datumaro.components.errors import DatumaroError


class Crypter:
    # Prefix (datum-) = 6 and Fernet = 44, 6 + 44 = 50
    FERNET_KEY_LEN = 50
    KEY_PREFIX = b"datum-"
    KEY_PREFIX_LEN = len(KEY_PREFIX)

    def __init__(self, key: Union[str, bytes]) -> None:
        if isinstance(key, str):
            key = key.encode()
        if len(key) != self.FERNET_KEY_LEN:
            raise DatumaroError(
                f"Key length should be {self.FERNET_KEY_LEN}, "
                f"but your key length is {len(key)} (key={key})."
            )
        self._key = key[self.KEY_PREFIX_LEN :]
        self._fernet = Fernet(self._key)

    @property
    def key(self) -> bytes:
        return self.KEY_PREFIX + self._key

    def decrypt(self, msg: bytes) -> bytes:
        return self._fernet.decrypt(msg)

    def encrypt(self, msg: bytes) -> bytes:
        return self._fernet.encrypt(msg)

    def handshake(self, key: bytes) -> bool:
        try:
            return self.decrypt(key) == self.key
        except InvalidToken as e:
            log.debug(e)
            return False

    @classmethod
    def gen_key(cls, key: Optional[bytes] = None) -> bytes:
        """If "key" is not None, return the different key with "key"."""
        _key = cls.KEY_PREFIX + Fernet.generate_key()
        while _key == key:
            _key = cls.KEY_PREFIX + Fernet.generate_key()
        return _key

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, Crypter):
            return False

        return self._key == __o._key

    @property
    def is_null_crypter(self):
        return self == NULL_CRYPTER


class NullCrypter(Crypter):
    def __init__(self) -> None:
        self._key = None
        self._fernet = None

    def decrypt(self, msg: bytes) -> bytes:
        return msg

    def encrypt(self, msg: bytes) -> bytes:
        return msg

    def handshake(self, key: bytes) -> bool:
        return self.decrypt(key) == b""

    @property
    def key(self) -> None:
        return None


NULL_CRYPTER = NullCrypter()



# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT


from datumaro.components.annotation import AnnotationType


class CvatPath:
    IMAGES_DIR = "images"

    IMAGE_EXT = ".jpg"

    BUILTIN_ATTRS = {"occluded", "outside", "keyframe", "track_id"}

    SUPPORTED_IMPORT_SHAPES = {
        "box",
        "polygon",
        "polyline",
        "points",
        "mask",
    }
    SUPPORTED_EXPORT_SHAPES = {
        AnnotationType.bbox,
        AnnotationType.polygon,
        AnnotationType.polyline,
        AnnotationType.points,
        AnnotationType.mask,
    }



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import os.path as osp
from collections import OrderedDict
from typing import Dict, Iterator, List, Optional, Type, TypeVar, Union

import yaml

from datumaro.components.annotation import Annotation, AnnotationType, Bbox, LabelCategories
from datumaro.components.dataset_base import CategoriesInfo, DatasetBase, DatasetItem, SubsetBase
from datumaro.components.errors import (
    DatasetImportError,
    InvalidAnnotationError,
    UndeclaredLabelError,
)
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image, ImageFromFile
from datumaro.util.image import (
    DEFAULT_IMAGE_META_FILE_NAME,
    IMAGE_EXTENSIONS,
    ImageMeta,
    load_image_meta_file,
)
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file
from datumaro.util.os_util import extract_subset_name_from_parent, find_files, split_path

from .format import YoloLoosePath, YoloPath, YoloUltralyticsPath

T = TypeVar("T")

__all__ = ["YoloStrictBase", "YoloLooseBase", "YoloUltralyticsBase"]


def localize_path(path: str) -> str:
    """
    Removes the "data/" prefix from the path
    """

    path = osp.normpath(path.strip()).replace("\\", "/")
    default_base = "data/"
    if path.startswith(default_base):
        path = path[len(default_base) :]
    return path


# TODO: Need to clean this dirty structure (SubsetBase has _Subset).
class _Subset(DatasetBase):
    def __init__(
        self,
        subset_name: str,
        path: str,
        items: OrderedDict[str, str],
        categories: CategoriesInfo,
        image_info: ImageMeta,
    ):
        super().__init__()
        self._subset_name = subset_name
        self._path = path
        self._items = items
        self._categories = categories
        self._image_info = image_info

    def __iter__(self) -> Iterator[DatasetItem]:
        for item_id in self._items:
            item = self._get(item_id)
            if item is not None:
                yield item

    def __len__(self):
        return len(self._items)

    def categories(self):
        return self._parent.categories()

    def _get(self, item_id: str) -> Optional[DatasetItem]:
        item = self._items.get(item_id)

        if not isinstance(item, str):
            return None

        try:
            image_size = self._image_info.get(item_id)
            image = Image.from_file(path=osp.join(self._path, item), size=image_size)

            anno_path = osp.splitext(image.path)[0] + ".txt"
            annotations = self._parse_annotations(
                anno_path,
                image,
                label_categories=self._categories[AnnotationType.label],
            )

            item = DatasetItem(
                id=item_id, subset=self._subset_name, media=image, annotations=annotations
            )
            return item
        except (UndeclaredLabelError, InvalidAnnotationError) as e:
            self._ctx.error_policy.report_annotation_error(e, item_id=(item_id, self._subset_name))
        except Exception as e:
            self._ctx.error_policy.report_item_error(e, item_id=(item_id, self._subset_name))

        return None

    @classmethod
    def _parse_annotations(
        cls,
        anno_path: str,
        image: ImageFromFile,
        *,
        label_categories: LabelCategories,
    ) -> List[Annotation]:
        lines = []
        with open(anno_path, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if line:
                    lines.append(line)

        annotations = []

        if lines:
            # Use image info as late as possible to avoid unnecessary image loading
            if image.size is None:
                raise DatasetImportError(f"Can't find image info for '{localize_path(image.path)}'")
            image_height, image_width = image.size

        for idx, line in enumerate(lines):
            parts = line.split()
            if len(parts) != 5:
                raise InvalidAnnotationError(
                    f"Unexpected field count {len(parts)} in the bbox description. "
                    "Expected 5 fields (label, xc, yc, w, h)."
                )
            label_id, xc, yc, w, h = parts

            label_id = cls._parse_field(label_id, int, "bbox label id")
            if label_id not in label_categories:
                raise UndeclaredLabelError(str(label_id))

            w = cls._parse_field(w, float, "bbox width")
            h = cls._parse_field(h, float, "bbox height")
            x = cls._parse_field(xc, float, "bbox center x") - w * 0.5
            y = cls._parse_field(yc, float, "bbox center y") - h * 0.5
            annotations.append(
                Bbox(
                    x * image_width,
                    y * image_height,
                    w * image_width,
                    h * image_height,
                    label=label_id,
                    id=idx,
                    group=idx,
                )
            )

        return annotations

    @classmethod
    def _parse_field(cls, value: str, desired_type: Type[T], field_name: str) -> T:
        try:
            return desired_type(value)
        except Exception as e:
            raise InvalidAnnotationError(
                f"Can't parse {field_name} from '{value}'. Expected {desired_type}"
            ) from e


class YoloStrictBase(SubsetBase):
    def __init__(
        self,
        config_path: str,
        image_info: Union[None, str, ImageMeta] = None,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
        **kwargs,
    ) -> None:
        super().__init__(subset=subset, ctx=ctx)

        if not osp.isfile(config_path):
            raise DatasetImportError(f"Can't read dataset descriptor file '{config_path}'")

        rootpath = osp.dirname(config_path)
        self._path = rootpath

        self._image_info = self.parse_image_info(rootpath, image_info)

        config = YoloPath._parse_config(config_path)

        names_path = config.get("names")
        if not names_path:
            raise InvalidAnnotationError("Failed to parse names file path from config")

        self._categories = {
            AnnotationType.label: self._load_categories(
                osp.join(self._path, localize_path(names_path))
            )
        }

        # The original format is like this:
        #
        # classes = 2
        # train  = data/train.txt
        # valid  = data/test.txt
        # names = data/obj.names
        # backup = backup/
        #
        # To support more subset names, we disallow subsets
        # called 'classes', 'names' and 'backup'.
        subsets = {k: v for k, v in config.items() if k not in YoloPath.RESERVED_CONFIG_KEYS}
        self._subsets: Dict[str, self._Subset] = {}

        for subset_name, list_path in subsets.items():
            if subset_name != subset:
                continue

            list_path = osp.join(self._path, localize_path(list_path))
            if not osp.isfile(list_path):
                raise InvalidAnnotationError(f"Can't find '{subset_name}' subset list file")

            with open(list_path, "r", encoding="utf-8") as f:
                items = OrderedDict(
                    (self.name_from_path(p), localize_path(p)) for p in f if p.strip()
                )

            subset = _Subset(
                subset_name=subset_name,
                path=self._path,
                items=items,
                categories=self._categories,
                image_info=self._image_info,
            )
            self._subsets[subset_name] = subset

    @staticmethod
    def parse_image_info(
        rootpath: str, image_info: Optional[Union[str, ImageMeta]] = None
    ) -> ImageMeta:
        assert image_info is None or isinstance(image_info, (str, dict))
        if image_info is None:
            image_info = osp.join(rootpath, DEFAULT_IMAGE_META_FILE_NAME)
            if not osp.isfile(image_info):
                image_info = {}
        if isinstance(image_info, str):
            image_info = load_image_meta_file(image_info)

        return image_info

    @classmethod
    def name_from_path(cls, path: str) -> str:
        """
        Obtains <image name> from the path like [data/]<subset>_obj/<image_name>.ext

        <image name> can be <a/b/c/filename>, so it is
        more involved than just calling "basename()".
        """

        path = localize_path(path)

        parts = split_path(path)
        if 1 < len(parts) and not osp.isabs(path):
            path = osp.join(*parts[1:])  # pylint: disable=no-value-for-parameter

        return osp.splitext(path)[0]

    @staticmethod
    def _load_categories(names_path):
        if has_meta_file(osp.dirname(names_path)):
            return LabelCategories.from_iterable(parse_meta_file(osp.dirname(names_path)).keys())

        label_categories = LabelCategories()

        with open(names_path, "r", encoding="utf-8") as f:
            for label in f:
                label = label.strip()
                if label:
                    label_categories.add(label)

        return label_categories

    def __iter__(self) -> Iterator[DatasetItem]:
        subsets = self._subsets
        pbars = self._ctx.progress_reporter.split(len(subsets))
        for pbar, (subset_name, subset) in zip(pbars, subsets.items()):
            for item in pbar.iter(subset, desc=f"Importing '{subset_name}'"):
                yield item

                for ann in item.annotations:
                    self._ann_types.add(ann.type)

    def __len__(self):
        return sum(len(s) for s in self._subsets.values())

    def get_subset(self, name):
        return self._subsets[name]

    @property
    def is_stream(self) -> bool:
        return True


class YoloLooseBase(SubsetBase):
    META_FILE = YoloLoosePath.NAMES_FILE

    def __init__(
        self,
        config_path: str,
        image_info: Union[None, str, ImageMeta] = None,
        urls: Optional[List[str]] = None,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ) -> None:
        super().__init__(subset=subset, ctx=ctx)

        if not osp.isdir(config_path):
            raise DatasetImportError(f"{config_path} should be a directory.")

        if not urls:
            raise DatasetImportError(
                f"`urls` should be specified for {self.__class__.__name__}, "
                f"if you want to import a dataset with using this {self.__class__.__name__} directly. "
                "In most case, it happens by giving an incorrect format name to the import interface. "
                "Please consider to import your dataset with this format name, 'yolo', "
                "such as `Dataset.import_from(..., format='yolo')`."
            )

        rootpath = self._get_rootpath(config_path)

        self._image_info = YoloStrictBase.parse_image_info(rootpath, image_info)

        # Init label categories
        label_categories = self._load_categories(osp.join(rootpath, self.META_FILE))
        self._categories = {AnnotationType.label: label_categories}

        self._urls = urls
        self._img_files = self._load_img_files(rootpath)

    def __iter__(self) -> Iterator[DatasetItem]:
        label_categories = self._categories.get(AnnotationType.label)
        if label_categories is None:
            raise DatasetImportError("label_categories should be not None.")

        pbar = self._ctx.progress_reporter
        for url in pbar.iter(self._urls, desc=f"Importing '{self._subset}'"):
            try:
                fname = self._get_fname(url)
                img = Image.from_file(path=self._img_files[fname])
                anns = self._parse_annotations(
                    url,
                    img,
                    label_categories=label_categories,
                )
                yield DatasetItem(id=fname, subset=self._subset, media=img, annotations=anns)

                for ann in anns:
                    self._ann_types.add(ann.type)
            except Exception as e:
                self._ctx.error_policy.report_item_error(e, item_id=(fname, self._subset))

    def __len__(self) -> int:
        return len(self._urls)

    def _get_rootpath(self, config_path: str) -> str:
        return config_path

    def _load_categories(self, names_path: str) -> LabelCategories:
        return YoloStrictBase._load_categories(names_path)

    def _get_fname(self, fpath: str) -> str:
        return osp.splitext(osp.basename(fpath))[0]

    def _load_img_files(self, rootpath: str) -> Dict[str, str]:
        return {
            self._get_fname(img_file): img_file
            for img_file in find_files(rootpath, IMAGE_EXTENSIONS, recursive=True, max_depth=2)
            if extract_subset_name_from_parent(img_file, rootpath) == self._subset
        }

    def _parse_annotations(
        self, anno_path: str, image: ImageFromFile, label_categories: LabelCategories
    ) -> List[Annotation]:
        return _Subset._parse_annotations(
            anno_path=anno_path, image=image, label_categories=label_categories
        )

    def _parse_field(self, value: str, desired_type: Type[T], field_name: str) -> T:
        return _Subset._parse_field(value=value, desired_type=desired_type, field_name=field_name)

    @property
    def is_stream(self) -> bool:
        return True


class YoloUltralyticsBase(YoloLooseBase):
    META_FILE = YoloUltralyticsPath.META_FILE

    def __init__(
        self,
        config_path: str,
        image_info: Union[None, str, ImageMeta] = None,
        urls: Optional[List[str]] = None,
        **kwargs,
    ) -> None:
        super().__init__(config_path, image_info, urls, **kwargs)

    def _load_categories(self, names_path: str) -> LabelCategories:
        if has_meta_file(osp.dirname(names_path)):
            return LabelCategories.from_iterable(parse_meta_file(osp.dirname(names_path)).keys())

        label_categories = LabelCategories()

        with open(names_path, "r") as fp:
            loaded = yaml.safe_load(fp.read())
            if isinstance(loaded["names"], list):
                label_names = loaded["names"]
            elif isinstance(loaded["names"], dict):
                label_names = list(loaded["names"].values())
            else:
                raise DatasetImportError(f"Can't read dataset category file '{names_path}'")

        for label_name in label_names:
            label_categories.add(label_name)

        return label_categories



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Optional, Sequence, Union

import numpy as np

from datumaro.components.algorithms.hash_key_inference.base import HashInference
from datumaro.components.algorithms.hash_key_inference.hashkey_util import (
    calculate_hamming,
    select_uninferenced_dataset,
)
from datumaro.components.annotation import HashKey
from datumaro.components.dataset import Dataset
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import DatumaroError, MediaTypeError


class Explorer(HashInference):
    def __init__(
        self,
        *datasets: Sequence[Dataset],
        topk: int = 10,
    ) -> None:
        """
        Explorer for Datumaro dataitems

        Parameters
        ----------
        dataset:
            Datumaro dataset to explore similar dataitem.
        topk:
            Number of images.
        """
        self._model = None
        self._text_model = None
        self._topk = topk
        database_keys = []
        item_list = []

        datasets_to_infer = [select_uninferenced_dataset(dataset) for dataset in datasets]
        datasets = self._compute_hash_key(datasets, datasets_to_infer)

        for dataset in datasets:
            for item in dataset:
                for annotation in item.annotations:
                    if isinstance(annotation, HashKey):
                        try:
                            hash_key = annotation.hash_key
                            hash_key = np.unpackbits(hash_key, axis=-1)
                            database_keys.append(hash_key)
                            item_list.append(item)
                        except Exception:
                            continue

        if all(i is None for i in database_keys):
            # media.data is None case
            raise ValueError("Database should have hash_key")

        self._database_keys = np.stack(database_keys, axis=0)
        self._item_list = item_list

    def explore_topk(
        self,
        query: Union[DatasetItem, str, List[Union[DatasetItem, str]]],
        topk: Optional[int] = None,
    ):
        """
        Explore topk similar results based on hamming distance for query DatasetItem
        """
        if not topk:
            topk = self._topk

        database_keys = self._database_keys

        if isinstance(query, list):
            topk_for_query = int(topk // len(query)) * 2 if not len(query) == 1 else topk
            query_hash_key_list = []
            result_list = []
            logits_list = []
            for query_ in query:
                if isinstance(query_, DatasetItem):
                    query_key = self._get_hash_key_from_item_query(query_)
                    query_hash_key_list.append(query_key)
                elif isinstance(query_, str):
                    query_key = self._get_hash_key_from_text_query(query_)
                    query_hash_key_list.append(query_key)
                else:
                    raise MediaTypeError(
                        "Unexpected media type of query '%s'. "
                        "Expected 'DatasetItem' or 'string', actual'%s'" % (query_, type(query_))
                    )

            for query_key in query_hash_key_list:
                unpacked_key = np.unpackbits(query_key.hash_key, axis=-1)
                logits = calculate_hamming(unpacked_key, database_keys)
                ind = np.argsort(logits)

                item_list = np.array(self._item_list)[ind]
                result_list.append(item_list[:topk_for_query].tolist())
                logits_list.append(logits[ind][:topk_for_query].tolist())

            result_list = np.stack(result_list, axis=0)
            logits_list = np.stack(logits_list, axis=0)

            flattened_indices = np.argsort(logits_list.ravel())
            sorted_list = result_list.ravel()[flattened_indices]

            return sorted_list[:topk]

        if isinstance(query, DatasetItem):
            query_key = self._get_hash_key_from_item_query(query)
        elif isinstance(query, str):
            query_key = self._get_hash_key_from_text_query(query)
        else:
            raise MediaTypeError(
                "Unexpected media type of query '%s'. "
                "Expected 'DatasetItem' or 'string', actual'%s'" % (query, type(query))
            )

        if not isinstance(query_key, HashKey):
            # media.data is None case
            raise ValueError("Query should have hash_key")

        unpacked_key = np.unpackbits(query_key.hash_key, axis=-1)
        logits = calculate_hamming(unpacked_key, database_keys)
        ind = np.argsort(logits)

        item_list = np.array(self._item_list)[ind]
        result = item_list[:topk].tolist()

        return result

    def _get_hash_key_from_item_query(self, query: DatasetItem) -> HashKey:
        """Get hash key from the `DatasetItem`.

        If not exists, launch the model inference to obtain it.
        """
        query_keys_in_item = [
            annotation for annotation in query.annotations if isinstance(annotation, HashKey)
        ]

        if len(query_keys_in_item) > 1:
            raise DatumaroError(
                f"There are more than two HashKey ({query_keys_in_item}) "
                f"in the query item ({query}). It is ambiguous!"
            )

        if len(query_keys_in_item) == 1:
            return query_keys_in_item[0]

        return self._model.infer_item(query)

    def _get_hash_key_from_text_query(self, query: str) -> HashKey:
        return self.text_model.infer_text(query)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from pathlib import Path

from datumaro.components.lazy_plugin import get_extra_deps, get_lazy_plugin
from datumaro.util import parse_json_file

_SOURCE_PATH = Path(__file__).resolve()
_SOURCE_DIR = _SOURCE_PATH.parent
_SPECS_JSON_PATH = _SOURCE_DIR / "specs.json"


def get_lazy_plugins():
    return [
        plugin
        for plugin in [
            get_lazy_plugin(
                spec["import_path"],
                spec["plugin_name"],
                spec["plugin_type"],
                spec.get("extra_deps", []),
                spec.get("metadata", {}),
            )
            for spec in parse_json_file(str(_SPECS_JSON_PATH))
        ]
        if plugin is not None
    ]


if __name__ == "__main__":
    from datumaro.components.environment import Environment
    from datumaro.util import dump_json_file

    env = Environment(use_lazy_import=False)
    plugin_specs = []

    def _enroll_to_plugin_specs(plugins, plugin_type):
        global plugin_specs

        for _, plugin in plugins.items():
            mod = plugin.__module__
            class_name = plugin.__name__
            plugin_name = plugin.NAME
            extra_deps = get_extra_deps(plugin)
            plugin_spec = {
                "import_path": f"{mod}.{class_name}",
                "plugin_name": plugin_name,
                "plugin_type": plugin_type,
            }
            if extra_deps:
                plugin_spec["extra_deps"] = extra_deps
            # Setting metadata for importers
            if hasattr(plugin, "get_file_extensions"):
                plugin_spec["metadata"] = {"file_extensions": sorted(plugin.get_file_extensions())}
            plugin_specs.append(plugin_spec)

    _enroll_to_plugin_specs(env.extractors, "DatasetBase")
    _enroll_to_plugin_specs(env.importers, "Importer")
    _enroll_to_plugin_specs(env.launchers, "Launcher")
    _enroll_to_plugin_specs(env.exporters, "Exporter")
    _enroll_to_plugin_specs(env.generators, "DatasetGenerator")
    _enroll_to_plugin_specs(env.transforms, "Transform")
    _enroll_to_plugin_specs(env.validators, "Validator")

    dump_json_file(
        _SPECS_JSON_PATH,
        sorted(plugin_specs, key=lambda spec: spec["import_path"]),
        indent=True,
        append_newline=True,
    )
